#!/usr/bin/env python
# filename: bin.py
#
#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import platform
from typing import Optional
from urllib.request import urlretrieve

__all__ = ["get_path", "get_binary_directory"]

# __all__ = [
#     "cdhit",
#     "fastp",
#     "fasttree",
#     "mmseqs",
#     "mafft",
#     "muscle",
#     "muscle_v3",
#     "vsearch",
# ]


BIN_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "binaries")
SYSTEM = platform.system().lower()
MACHINE = platform.machine().lower().replace("x86_64", "amd64")

# cdhit = os.path.join(BIN_DIR, f"cd-hit_{SYSTEM}_{MACHINE}")
# fastp = os.path.join(BIN_DIR, f"fastp_{SYSTEM}")
# fasttree = os.path.join(BIN_DIR, f"fasttree_{SYSTEM}_{MACHINE}")
# mafft = os.path.join(BIN_DIR, f"mafft_{SYSTEM}_{MACHINE}/mafft.bat")
# mmseqs = os.path.join(BIN_DIR, f"mmseqs_{SYSTEM}_{MACHINE}")
# muscle = os.path.join(BIN_DIR, f"muscle_{SYSTEM}_{MACHINE}")
# muscle_v3 = os.path.join(BIN_DIR, f"muscle3_{SYSTEM}_{MACHINE}")
# vsearch = os.path.join(BIN_DIR, f"vsearch_{SYSTEM}_{MACHINE}")


def get_binary_directory() -> str:
    """
    Gets the abutils binary directory.

    Returns
    -------
    str
        Path to the abutils binary directory.

    """
    return BIN_DIR


def copy_to_binary_directory(source: str, name: Optional[str] = None) -> None:
    """
    Copies a binary to the abutils binary directory.

    Parameters
    ----------
    source : str
        Path to the binary to be copied.

    name : str, optional
        Name of the binary in the binary directory. If not provided, the basename of the
        source binary will be used.

    """
    binary_dir = get_binary_directory()
    name = name if name is not None else os.path.basename(source)
    target = os.path.join(binary_dir, source)
    if not os.path.exists(target):
        os.system(f'cp "{source}" "{target}"')


def get_path(binary: str) -> str:
    """
    Gets the path to a pre-packaged binary.

    .. note::
        If the binary is not found in the binary directory, it will be downloaded from
        the brineylab S3 bucket. This is because some binaries are too large
        to be included in the ``abutils`` PyPI package.

    Parameters
    ----------
    binary : str
        Name of the binary. Available binaries are:
            - blastn
            - cd-hit
            - fastp
            - fasttree
            - makeblastdb
            - mafft
            - minimap2
            - mmseqs
            - muscle
            - muscle_v3
            - sickle
            - vsearch

    Returns
    -------
    str
        Path to the binary.

    """
    #  format (and maybe fix) binary name
    binary = binary.lower()
    if binary == "mmseqs2":
        binary = "mmseqs"
    if binary == "muscle_v3":
        binary = "muscle3"
    if binary == "cd-hit":
        binary = "cdhit"
    if binary == "minimap":
        binary = "minimap2"
    available = [
        # blast
        "blastn",
        "makeblastdb",
        # clustering/alignment
        "cdhit",
        "mafft",
        "minimap2",
        "mmseqs",
        "muscle",
        "muscle3",
        "vsearch",
        # qc/preprocessing
        "fastp",
        "sickle",
        # phylogeny
        "fasttree",
    ]
    if binary not in available:
        raise ValueError(
            f"Binary '{binary}' not available. Available binaries: \n"
            + "\n".join(available)
        )
    # get binary path
    if binary in ["fastp", "muscle3", "blastn", "makeblastdb"]:
        bin_name = f"{binary}_{SYSTEM}"
        bin_path = os.path.join(BIN_DIR, bin_name)
    elif binary == "mafft":
        bin_name = f"{binary}_{SYSTEM}/mafft.bat"
        bin_path = os.path.join(BIN_DIR, bin_name)
    else:
        bin_name = f"{binary}_{SYSTEM}_{MACHINE}"
        bin_path = os.path.join(BIN_DIR, bin_name)
    # download binary if missing
    if not os.path.exists(bin_path):
        print(
            f"Downloading missing binary: {binary}. This will only occur once on intial use of the binary."
        )
        download_missing_binary(bin_name, bin_path)
    return bin_path


# download missing binary (some are too large to included in the PyPI package)
def download_missing_binary(bin_name: str, bin_path: str) -> None:
    url = f"https://brineylab.s3.amazonaws.com/tools/abutils_binaries/{bin_name}"
    urlretrieve(url, bin_path)


# # download any missing binaries (some are too large to included in the PyPI package)
# def download_missing_binaries():
#     to_download = []
#     for b in [
#         "cd-hit",
#         "fastp",
#         "fasttree",
#         "mafft",
#         "mmseqs",
#         "muscle",
#         "muscle3",
#         "vsearch",
#     ]:
#         if b in ["fastp", "muscle3"]:
#             bin_name = f"{b}_{SYSTEM}"
#             bin_path = os.path.join(BIN_DIR, bin_name)
#         else:
#             bin_name = f"{b}_{SYSTEM}_{MACHINE}"
#             bin_path = os.path.join(BIN_DIR, bin_name)
#         if not os.path.exists(bin_path):
#             to_download.append((bin_name, bin_path))
#     if to_download:
#         print(
#             "Downloading missing binaries. This will only occur once on intial use of abutils."
#         )
#         url = f"https://brineylab.s3.amazonaws.com/tools/abutils_binaries/{bin_name}"
#         urlretrieve(url, bin_path)


# download_missing_binaries()



#!/usr/bin/env python
# filename: color.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from typing import Iterable, Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt

__all__ = [
    "get_cmap",
    "palettes",
    "true_false_palette",
    "show_palettes",
    "hex_to_rgb",
    "rgb_to_hex",
    "hls",
    "husl",
    "truncate_colormap",
    "monochrome_palette",
]


# -----------------
#    PALETTES
# -----------------

palettes = {
    "muted_neon": [
        "#ef476f",
        "#ffd166",
        "#06d6a0",
        "#118ab2",
        "#6c6678",
        "#073b4c",
    ],
    "pastel": ["#9b5de5", "#f15bb5", "#fee440", "#00bbf9", "#00f5d4"],
    "sunset": ["#264653", "#40768c", "#2a9d8f", "#e9c46a", "#f4a261", "#e76f51"],
    "vibrant": ["#ff595e", "#ffca3a", "#8ac926", "#1982c4", "#6a4c93"],
    "bright_pastel": ["#826aed", "#c879ff", "#ffb7ff", "#3bf4fb", "#caff8a"],
    "fresh": ["#1be7ff", "#6eeb83", "#e4ff1a", "#ffb800", "#ff5714"],
    "cool": [
        "#065275",
        "#029099",
        "#7ccba1",
        "#fdde9c",
        "#f0746e",
        "#dc3978",
        "#7c1e6f",
    ],
    "primary": ["#246eb9", "#4cb944", "#c73e1d", "#ffbe0b", "#401f3e"],
}

true_false_palette = {True: "#e41a1c", False: "#d1d1d1"}


def show_palettes() -> None:
    """
    Displays all abutils built-in color palettes. Inspired by matplotlib's `colormaps`_ tutorial.

    .. _colormaps: https://matplotlib.org/stable/tutorials/colors/colormaps.html
    """
    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient, gradient))

    nrows = len(palettes)
    figh = 1 + (nrows + (nrows - 1) * 0.1) * 0.75
    fig, axs = plt.subplots(
        nrows=nrows + 1,
        figsize=(6.4, figh),
    )
    fig.subplots_adjust(
        top=1 - 0.35 / figh, bottom=0.15 / figh, left=0.2, right=0.99, hspace=0.6
    )
    for ax, name in zip(axs, palettes):
        cmap = mpl.colors.ListedColormap(palettes[name])
        ax.imshow(
            gradient,
            aspect="auto",
            cmap=cmap,
        )
        ax.text(
            -0.02,
            0.5,
            name,
            va="center",
            ha="right",
            fontsize=10,
            transform=ax.transAxes,
        )

    for ax in axs:
        ax.set_axis_off()

    plt.show()


def get_cmap(
    c: Union[mpl.colors.Colormap, str, Iterable],
    dark: bool = False,
    zero_color: Union[str, Iterable, None] = None,
    n: int = 256,
    minval: float = 0.0,
    maxval: float = 1.0,
    name: Optional[str] = None,
) -> mpl.colors.Colormap:
    """
    Gets a matplotlib ``Colormap``.

    Parameters
    ----------
    c : ``str``, ``Colormap`` or ``tuple``
        Can be one of several things:
            * matplotlib ``Colormap``
            * the name of a matplotlib ``Colormap`` (e.g. ``'viridis'``)
            * hex code
            * a `matplotlib color`_
            * RGB tuple
        If a single color is provided, a ``Colormap`` will be built from
        white to the provided color (or from the color to black if `dark`
        is ``True``).

    dark : bool, default=False
        If ``True``, build a ``Colormap`` from the provided color from the color
        to black. By default, the ``Colormap`` will be built from the provided
        color to white.

    zero_color : any Matplotlib color, default=None
        Alternate color for the zero value of the ``Colormap``. Produces results
        similar to the default colormap in 10x Genomics' Loupe browser, in which
        the zero count value for GEX plots is light grey and less easily confused with
        near-zero counts.

    minval : float, default=0.0
        Truncate the lower bound of the ``Colormap``. Must be a ``float`` less than 1.0,
        which represents the fraction of the ``Colormap`` at which to truncate.

    maxval : float, default=1.0
        Truncate the upper bound of the ``Colormap``. Must be a ``float`` greater than
        zero and less than or equal to 1.0, which represents the fraction of the
        ``Colormap`` to truncate.

    n : int, default=256
        Number of steps in the ``Colormap``.

    name : str, default=None
        Name of the ``Colormap``.


    Returns
    -------
    cmap : ``Colormap``


    .. _Matplotlib color
        https://matplotlib.org/stable/tutorials/colors/colors.html
    """
    if isinstance(c, mpl.colors.Colormap):
        cmap = c
    elif isinstance(c, str) and c in plt.colormaps():
        cmap = plt.get_cmap(c)
    elif dark:
        cmap = sns.dark_palette(c, as_cmap=True)
    else:
        cmap = sns.light_palette(c, as_cmap=True)
    if zero_color is not None:
        cmap = cmap(np.linspace(0.1, 1, n))
        colors = [mpl.colors.to_rgba_array(zero_color)] + list(cmap)
        cmap = mpl.colors.LinearSegmentedColormap.from_list(
            name if name is not None else "", colors
        )
    cmap = mpl.colors.LinearSegmentedColormap.from_list(
        name, cmap(np.linspace(minval, maxval, n))
    )
    return cmap


def monochrome_palette(
    color: Union[str, Iterable], n_colors: int = 10, include_white: bool = False
) -> list:
    """
    Returns a monochromatic palette of colors, from `color` to white.

    Parameters
    ----------
    color : str or Iterable
        Color from which the monochromatic palette will be created.
        Can be one of several things:
            * hex code
            * `Matplotlib color`_
            * RGB tuple

    n_colors : int, default=10
        Number of colors in the palette.

    include_white : bool, default=False
        Whether or not to include white in the palette. White is not
        included by default, meaning all colors in the palette will be
        a shade of `color`.


    Returns
    -------
    palette : list
        A ``list`` of RGBA ``tuple``s


    .. _Matplotlib color
        https://matplotlib.org/stable/tutorials/colors/colors.html
    """
    cmap = get_cmap(color)
    if include_white:
        colors = [cmap(i) for i in np.linspace(1, 0, n_colors)]
    else:
        colors = [cmap(i) for i in np.linspace(1, 0, n_colors + 1)][:-1]
    return colors


def cmap_from_color(
    color: Union[str, Tuple], dark: bool = False
) -> mpl.colors.Colormap:
    """
    Generates a matplotlib colormap from a single color. Colormap will be built,
    by default, from white to ``color``.

    Parameters
    ----------
    color : str or Iterable
        Color from which the monochromatic palette will be created.
        Can be one of several things:
            * hex code
            * `Matplotlib color`_
            * RGB tuple

    dark : bool, default=False
        If ``True``, colormap will be built from ``color`` to
        black. Default is ``False``, which builds a colormap from
        white to ``color``.


    Returns
    -------
        colormap : ``Colormap``


    .. _Matplotlib color
        https://matplotlib.org/stable/tutorials/colors/colors.html
    """
    return get_cmap(color, dark=dark)


# cmaps = {
#     "heatmap": sns.diverging_palette(240, 10, as_cmap=True),
#     "loupe": get_cmap("YlOrBr", zero_color=[0.9, 0.9, 0.9, 1.0], minval=0.1),
# }


def hex_to_rgb(hex_string):
    rgb = mpl.colors.hex2color(hex_string)
    return tuple([int(255 * x) for x in rgb])


def rgb_to_hex(rgb_tuple):
    div = 1 if all([v <= 1.0 for v in rgb_tuple]) else 255
    return mpl.colors.rgb2hex([1.0 * x / div for x in rgb_tuple])


def hls(n_colors, hue=0.01, lightness=0.6, saturation=0.65):
    return sns.hls_palette(n_colors, h=hue, l=lightness, s=saturation)


def husl(n_colors, hue=0.01, saturation=0.9, lightness=0.65):
    return sns.husl_palette(n_colors, h=hue, s=saturation, l=lightness)


def truncate_colormap(
    cmap: mpl.colors.Colormap, minval: float = 0.0, maxval: float = 1.0, n: int = 256
) -> mpl.colors.Colormap:
    """
    Truncates a colormap, such that the new colormap consists of
    ``cmap[minval:maxval]``.

    If maxval is larger than minval, the truncated colormap will be reversed.

    Args:

        cmap : colormap): Colormap to be truncated.

        minval (float): Lower bound. Should be a float betwee 0 and 1.

        maxval (float): Upper bound. Should be a float between 0 and 1

        n (int): Number of colormap steps. Default is ``256``.

    Returns:

        colormap: A matplotlib colormap

    http://stackoverflow.com/questions/18926031/how-to-extract-a-subset-of-a-colormap-as-a-new-colormap-in-matplotlib
    """
    # cmap = get_cmap(cmap)
    name = "%s-trunc-%.2g-%.2g" % (cmap.name, minval, maxval)
    return mpl.colors.LinearSegmentedColormap.from_list(
        name, cmap(np.linspace(minval, maxval, n)), N=n
    )


# def stack_colormap(lower, upper, n=256):
#     """
#     Stacks two colormaps (``lower`` and ``upper``) such that
#     low half -> ``lower`` colors, high half -> ``upper`` colors

#     Args:

#     	lower (colormap): colormap for the lower half of the stacked colormap.

#     	upper (colormap): colormap for the upper half of the stacked colormap.

#     	n (int): Number of colormap steps. Default is ``256``.
#     """
#     A = get_cmap(lower)
#     B = get_cmap(upper)
#     name = "%s-%s" % (A.name, B.name)
#     lin = np.linspace(0, 1, n)
#     return array_cmap(np.vstack((A(lin), B(lin))), name, n=n)


# def get_cmap(cmap=None, name=None, from_color=None, dark=False, n=256):
#     # """
#     # Generates a matplotlib colormap.

#     # cmap can be one of several things:
#     #     - a name ('Blues', 'BuGn_r') of a built-in colormap
#     #     - a cmap
#     #     - a filename, np.loadtxt() n x 3 or 4  ints 0..255 or floats 0..1
#     #     - a numpy array
#     # See http://wiki.scipy.org/Cookbook/Matplotlib/Show_colormaps or in IPython, plt.cm.<tab>

#     # An optional name for the colormap can be provided (name), as well as the number
#     # of cmap steps (n).

#     # Alternatively, to make a colormap using a single color, provide the color
#     # via from_color. By default, the supplied color will be the dark end of
#     # the cmap, with white as the lightest color. To reverse (use the input
#     # color as the lighest value and black as the darkest), set dark=True.

#     # Returns a matplotlib colormap object.
#     # """
#     if from_color is not None:
#         return cmap_from_color(from_color, dark)
#     elif cmap is None:
#         err = "You must provide either cmap or from_color"
#         raise RuntimeError(err)
#     if isinstance(cmap, colors.Colormap):
#         return cmap
#     if isinstance(cmap, str):
#         if cmap in cm.cmap_d:
#             return plt.get_cmap(cmap)  # "Blues" ...
#         A = np.loadtxt(cmap, delimiter=None)  # None: white space
#         name = name or cmap.split("/")[-1].split(".")[0]  # .../xx.csv -> xx
#     else:
#         A = cmap  # numpy array or array-like
#     return array_cmap(A, name, n=n)


# def array_cmap(A, name=None, n=256):
#     """ numpy array -> a cmap, matplotlib.colors.Colormap
#         n x 3 or 4  ints 0 .. 255 or floats 0 ..1
#     """
#     A = np.asanyarray(A)
#     assert A.ndim == 2 and A.shape[1] in (
#         3,
#         4,
#     ), "array must be n x 3 or 4, not %s" % str(A.shape)
#     Amin, Amax = A.min(), A.max()
#     if A.dtype.kind == "i":
#         assert 0 <= Amin < Amax <= 255, "Amin %d  Amax %d must be in 0 .. 255" % (
#             Amin,
#             Amax,
#         )
#         A = A / 255.0  # not /=
#     else:
#         assert 0 <= Amin < Amax <= 1, "Amin %g  Amax %g must be in 0 .. 1" % (
#             Amin,
#             Amax,
#         )
#     return colors.LinearSegmentedColormap.from_list(name or "noname", A, N=n)



#!/usr/bin/env python
# filename: seqio.py


#
# Copyright (c) 2018 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# import json
# import os
# import sys
# from abc import ABCMeta, abstractmethod

# from Bio import SeqIO

# from ..core.sequence import Sequence
# from ..io import list_files
# from . import mongodb
# from .decorators import lazy_property

# # if sys.version_info[0] > 2:
# #     STR_TYPES = [
# #         str,
# #     ]
# # else:
# #     STR_TYPES = [str, unicode]


# # def read_input(input, data_type,
# #                collection=None, mongo_ip='localhost', mongo_port=27017, mongo_user=None, mongo_password=None,
# #                query=None, projection=None, verbose=False, **kwargs):
# #     '''
# #     Returns an Input class based on the data information provided.

# #     Args:

# #         data_type (str): One of the following: `'fasta'`, `'json'`, or `'mongodb'`.

# #         input (str): Path to an input file for FASTA or JSON data types, or the database name for MongoDB data.

# #         collection (str): Name of a MongoDB collection. Required for the MongoDB data type.

# #         mongo_ip (str): IP address of the MongoDB server. Default is `'localhost'` if not provided.

# #         mongo_port (int): Port of the MongoDB server. Default is `27017` if not provided.

# #         query (dict): Query to limit the results returned from a MongoDB database.

# #         projection (dict): Projection to specify fields to be retuired from a MongoDB database.
# #     '''
# #     if data_type.lower() == 'mongodb':
# #         return MongoDBInput(database=input, collection=collection, ip=mongo_ip, port=mongo_port,
# #                             user=mongo_user, password=mongo_password, query=query, projection=projection)
# #     elif data_type.lower() == 'json':
# #         return JSONInput(input)
# #     elif data_type.lower() == 'fasta':
# #         return FASTAInput(input)
# #     else:
# #         err = '\n\nERROR: data_type must be one of the following:\n'
# #         err += 'json, mongodb, or fasta\n\n'
# #         print(err)
# #         sys.exit(1)


# def read_fasta(fasta_file, verbose=False):
#     return FASTAInput(fasta_file, verbose=verbose)


# def from_fasta(fasta_file, verbose=False):
#     return FASTAInput(fasta_file, verbose=verbose)


# def from_json(json_file, seq_field="vdj_nt", verbose=False):
#     return JSONInput(json_file, seq_field=seq_field, verbose=verbose)


# def from_mongodb(
#     db,
#     collection=None,
#     ip="localhost",
#     port=27017,
#     user=None,
#     password=None,
#     query=None,
#     projection=None,
#     seq_field="vdj_nt",
#     verbose=False,
# ):
#     return MongoDBInput(
#         database=db,
#         collection=collection,
#         ip=ip,
#         port=port,
#         user=user,
#         password=password,
#         query=query,
#         projection=projection,
#         seq_field=seq_field,
#         verbose=verbose,
#     )


# class BaseInput:
#     """
#     Base class for parsing inputs (JSON, MongoDB, FASTA files, etc)
#     """

#     __metaclass__ = ABCMeta

#     def __init__(self):
#         pass

#     @property
#     @abstractmethod
#     def data_type(self):
#         "Returns the data type"
#         pass

#     @property
#     @abstractmethod
#     def as_list(self):
#         "Returns the input as a list of Sequence objects"
#         pass

#     @property
#     @abstractmethod
#     def as_generator(self):
#         "Returns the input as a genarator of Sequence objects"
#         pass


# class FASTAInput(BaseInput):
#     """
#     Representation of FASTA input data.
#     """

#     def __init__(self, data, verbose=False):
#         self.input = data
#         self.verbose = verbose

#     @property
#     def data_type(self):
#         return "fasta"

#     @property
#     def files(self):
#         if isinstance(self.input, str):
#             if os.path.isdir(self.input):
#                 return list_files(self.input, "json")
#             else:
#                 return [
#                     self.input,
#                 ]
#         else:
#             return self.input

#     @lazy_property
#     def as_list(self):
#         sequences = []
#         for input_file in self.files:
#             if self.verbose:
#                 print(input_file)
#             with open(input_file, "r") as f:
#                 for seq in SeqIO.parse(f, "fasta"):
#                     sequences.append(Sequence(str(seq.seq), id=seq.id))
#         return sequences

#     @property
#     def as_generator(self):
#         for input_file in self.files:
#             if self.verbose:
#                 print(input_file)
#             with open(input_file, "r") as f:
#                 for seq in SeqIO.parse(f, "fasta"):
#                     yield Sequence(str(seq.seq), id=seq.id)


# class JSONInput(BaseInput):
#     """
#     Representation of JSON input data
#     """

#     def __init__(self, data, seq_field="vdj_nt", verbose=False):
#         self.input = data
#         self.seq_field = seq_field
#         self.verbose = verbose

#     @property
#     def data_type(self):
#         return "json"

#     @property
#     def files(self):
#         if isinstance(self.input, str):
#             if os.path.isdir(self.input):
#                 return list_files(self.input, "json")
#             else:
#                 return [
#                     self.input,
#                 ]
#         else:
#             return self.input

#     @lazy_property
#     def as_list(self):
#         sequences = []
#         for input_file in self.files:
#             if self.verbose:
#                 print(input_file)
#             with open(input_file, "r") as f:
#                 for line in f:
#                     j = json.loads(line.strip().lstrip("]").rstrip("]").rstrip(","))
#                     sequences.append(Sequence(j, seq_key=self.seq_field))
#         return sequences

#     @property
#     def as_generator(self):
#         for input_file in self.files:
#             if self.verbose:
#                 print(input_file)
#             with open(input_file, "r") as f:
#                 for line in f:
#                     j = json.loads(
#                         line.strip().lstrip("[").rstrip("]").rstrip().rstrip(",")
#                     )
#                     yield Sequence(j, seq_key=self.seq_field)


# class MongoDBInput(BaseInput):
#     """
#     Representation of MongoDB input data
#     """

#     def __init__(
#         self,
#         database,
#         collection,
#         ip,
#         port,
#         user,
#         password,
#         query,
#         projection,
#         seq_field="vdj_nt",
#         verbose=False,
#     ):
#         self.db_name = database
#         self.raw_collections = collection
#         self.ip = ip
#         self.port = port
#         self.user = user
#         self.password = password
#         self.query = query
#         self.projection = projection
#         self.seq_field = seq_field
#         self.verbose = verbose

#     @property
#     def data_type(self):
#         return "mongodb"

#     @property
#     def db(self):
#         return mongodb.get_db(
#             self.db_name,
#             ip=self.ip,
#             port=self.port,
#             user=self.user,
#             password=self.password,
#         )

#     @property
#     def collections(self):
#         if isinstance(self.raw_collections, str):
#             return [
#                 self.raw_collections,
#             ]
#         elif self.raw_collections is None:
#             return mongodb.get_collections(self.db)
#         else:
#             return self.raw_collections

#     @lazy_property
#     def as_list(self):
#         sequences = []
#         for collection in self.collections:
#             if self.verbose:
#                 print(collection)
#             res = self.db[collection].find(self.query, self.projection)
#             for r in res:
#                 if self.seq_field not in r:
#                     continue
#                 sequences.append(Sequence(r, seq_key=self.seq_field))
#         return sequences

#     @property
#     def as_generator(self):
#         for collection in self.collections:
#             if self.verbose:
#                 print(collection)
#             res = self.db[collection].find(self.query, self.projection)
#             for r in res:
#                 if self.seq_field not in r:
#                     continue
#                 yield Sequence(r, seq_key=self.seq_field)

#     def _process_collections(self, collection):
#         if isinstance(collection, str):
#             return [
#                 collection,
#             ]
#         elif collection is None:
#             return mongodb.get_collections(self.db)
#         else:
#             return collection



#!/usr/bin/env python
# filename: phylogeny.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# from __future__ import absolute_import, division, print_function, unicode_literals

# import colorsys
# import itertools
# import math
# import multiprocessing as mp
# import os
# import platform
# import random
# import shutil
# import string
# import subprocess as sp
# import sys
# import tempfile
# from collections import Counter
# from copy import copy, deepcopy
# from typing import Optional

# # import abstar
# import ete3
# import matplotlib.pyplot as plt
# import numpy as np
# import pandas as pd
# import scipy
# import seaborn as sns

# # from abstar.core.germline import get_imgt_germlines
# from Bio import AlignIO, Phylo
# from matplotlib.colors import ListedColormap
# from scipy.ndimage import gaussian_filter1d
# from scipy.special import rel_entr

# from ..core.pair import Pair
# from ..core.sequence import Sequence
# from .alignment import mafft, muscle
# from .cluster import cluster
# from .color import get_cmap, hex_to_rgb
# from .decorators import lazy_property
# from .pipeline import make_dir

# if sys.version_info[0] > 2:
#     STR_TYPES = [
#         str,
#     ]
# else:
#     STR_TYPES = [str, unicode]


# # --------------------------------
# #  PHYLOGENETIC RECONSTRUCTION
# # --------------------------------


# def fasttree(
#     aln: str,
#     tree_file: Optional[str] = None,
#     is_aa: bool = False,
#     fasttree_bin: Optional[str] = None,
#     debug: bool = False,
#     quiet: bool = True,
# ) -> str:
#     """
#     Computes a tree file from a multiple seqeunce alignment using `FastTree`_.

#     Parameters
#     ----------
#     aln : str
#         Path to a multiple sequence alignment file, in FASTA format, or a
#         FASTA-formatted multiple sequence alignment string. Required.

#     tree_file : str
#         Path to the tree file which will be output by FastTree. If the parent
#         directory does not exist, it will be created. If not provided, the output
#         (a Newick-formatted tree file) will be returned as a ``str``.

#     is_aa : bool, default=False
#         Must be set to ``True`` if the input multiple sequence alignment contains
#         amino acid sequences. Default is ``False``, meaning FastTree will expect
#         nucleotide sequences.

#     fasttree_bin : str, optional
#         Path to the desired FastTree binary. Default is to use the version of
#         FastTree that is bundled with ``abutils``.

#     debug : bool, default=False
#         If ``True``, verbose output is printed. Default is False.

#     quiet : bool, default=True
#         Depricated, but retained for backwards compatibility. Use `debug` instead.


#     Returns
#     -------
#     tree_file: str
#         Path to the tree file produced by FastTree.


#     .. _FastTree:
#         http://www.microbesonline.org/fasttree/
#     """
#     # process input
#     if os.path.isfile(aln):
#         alignment_file = os.path.abspath(aln)
#     else:
#         ff = tempfile.NamedTemporaryFile(delete=False)
#         ff.close()
#         alignment_file = ff.name
#         with open(alignment_file, "w") as f:
#             f.write(aln)
#     if not quiet:
#         debug = True
#     # set the FastTree binary
#     if fasttree_bin is None:
#         mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#         fasttree_bin = os.path.join(
#             mod_dir, f"bin/fasttree_{platform.system().lower()}"
#         )
#     # make output directory if necessary
#     if tree_file is None:
#         as_file = False
#         tree_file = tempfile.NamedTemporaryFile(delete=False).name
#     else:
#         as_file = True
#         tree_file = os.path.abspath(tree_file)
#     if not os.path.isdir(os.path.dirname(tree_file)):
#         make_dir(os.path.dirname(tree_file))
#     # run FastTree
#     if is_aa:
#         ft_cmd = "fasttree {} > {}".format(alignment_file, tree_file)
#     else:
#         ft_cmd = "fasttree -nt {} > {}".format(alignment_file, tree_file)
#     ft = sp.Popen(ft_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = ft.communicate()
#     if debug:
#         print(ft_cmd)
#         print(stdout)
#         print(stderr)
#     # output
#     if as_file:
#         return tree_file
#     else:
#         with open(tree_file, "r") as f:
#             tree_string = f.read()
#         return tree_string


# def lsd(
#     tree,
#     output_file=None,
#     dates_file=None,
#     outgroup_file=None,
#     with_constraints=True,
#     with_weights=True,
#     reestimate_root_position=None,
#     quiet=True,
# ):
#     lsd_cmd = "lsd -i {}".format(os.path.abspath(tree))
#     if output_file is not None:
#         lsd_cmd += " -o {}".format(os.path.abspath(output_file))
#     if dates_file is not None:
#         lsd_cmd += " -d {}".format(os.path.abspath(dates_file))
#     if outgroup_file is not None:
#         lsd_cmd += " -g {}".format(os.path.abspath(outgroup_file))
#     if with_constraints:
#         lsd_cmd += " -c"
#     if with_weights:
#         lsd_cmd += " -v"
#     if reestimate_root_position is not None:
#         lsd_cmd += " -r {}".format(reestimate_root_position)
#     p = sp.Popen(lsd_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if not quiet:
#         print(lsd_cmd)
#         print(stdout)
#         print(stderr)
#     return output_file


# def igphyml(
#     input_file: str = None,
#     tree_file: str = None,
#     root: str = None,
#     verbose: bool = False,
# ) -> str:
#     """
#     Computes a phylogenetic tree using IgPhyML.

#     .. note::

#         IgPhyML must be installed. It can be downloaded from https://github.com/kbhoehn/IgPhyML.

#     Args:

#         input_file (str): Path to a Phylip-formatted multiple sequence alignment. Required.

#         tree_file (str): Path to the output tree file.

#         root (str): Name of the root sequence. Required.

#         verbose (bool): If `True`, prints the standard output and standard error for each IgPhyML run.
#             Default is `False`.
#     """

#     if shutil.which("igphyml") is None:
#         raise RuntimeError(
#             "It appears that IgPhyML is not installed.\nPlease install and try again."
#         )

#     # first, tree topology is estimated with the M0/GY94 model
#     igphyml_cmd1 = "igphyml -i {} -m GY -w M0 -t e --run_id gy94".format(aln_file)
#     p1 = sp.Popen(igphyml_cmd1, stdout=sp.PIPE, stderr=sp.PIPE)
#     stdout1, stderr1 = p1.communicate()
#     if verbose:
#         print(stdout1 + "\n")
#         print(stderr1 + "\n\n")
#     intermediate = input_file + "_igphyml_tree.txt_gy94"

#     # now  we fit the HLP17 model once the tree topology is fixed
#     igphyml_cmd2 = "igphyml -i {0} -m HLP17 --root {1} -o lr -u {}_igphyml_tree.txt_gy94 -o {}".format(
#         input_file, root, tree_file
#     )
#     p2 = sp.Popen(igphyml_cmd2, stdout=sp.PIPE, stderr=sp.PIPE)
#     stdout2, stderr2 = p2.communicate()
#     if verbose:
#         print(stdout2 + "\n")
#         print(stderr2 + "\n")
#     return tree_file + "_igphyml_tree.txt"


# # --------------------------------
# #      PHYLOGENETIC TREES
# # --------------------------------


# def phylogeny(
#     sequences=None,
#     project_dir=None,
#     name=None,
#     aln_file=None,
#     tree_file=None,
#     seq_field=None,
#     name_field=None,
#     aa=False,
#     species="human",
#     unrooted=False,
#     ladderize=True,
#     root=None,
#     root_name=None,
#     show_root_name=False,
#     color_dict=None,
#     color_function=None,
#     order_dict=None,
#     order_function=None,
#     color_node_labels=False,
#     label_colors=None,
#     scale=None,
#     branch_vert_margin=None,
#     fontsize=12,
#     show_names=True,
#     show_scale=False,
#     mirror=False,
#     min_order_fraction=0.1,
#     figname_prefix=None,
#     figname_suffix=None,
#     # linked_alignment=None, alignment_fontsize=11, alignment_height=50, alignment_width=50,
#     compact_alignment=False,
#     scale_factor=1,
#     rename_function=None,
#     linewidth=1.0,
#     delete_nodes=None,
#     quiet=True,
# ):
#     """
#     Generates a lineage phylogeny figure.

#     Args:

#         sequences (list(Sequence)): A list of ``Sequence`` objects from which a phylogeny
#             will be calculated. Strictly speaking, they do not need to be ``Sequence`` objects,
#             rather, any object that contains the sequence name as the ``id`` attribute (or
#             by dictionary-style lookup using the provided ``name_field``) and contains the
#             sequence as the ``sequence`` attribute (or by dictionary-stype lookup using the
#             provided ``seq_field``).

#         project_dir (str): directory into which all phylogeny files will be deposited,
#             including alignment, tree and figure files.

#         name (str): Name to be used for naming alignment, tree, and phylogeny files. If not
#             provided, a random name will be generated.

#         aln_file (str): if a multiple sequence alignment has already been calculated,
#             passing the path to the alignment file (in FASTA format) will force Lineage.phylogeny()
#             to use the supplied msa instead of computing a new one.

#         tree_file (str): if a tree file has already been calculated, passing the path
#             to the pre-computed tree file will force ``phylogeny()`` to use
#             the supplied tree file instead of computing a new one. It is important to note that
#             only sequence names will be parsed from the tree_file, so if ``order_function`` or
#             ``color_function`` is also provided, ensure that these functions only require the
#             sequence ID rather than the entire sequence.

#         aa (bool): if True, use amino acid sequences to compute the phylogeny.
#             Default is False.

#         root (Sequence, str: The root can be provided either as a ``Sequence`` object (if ``sequences``
#             are being provided) or as the name of a sequence that can be found either in
#             ``sequences`` or in the provided ``aln_file`` or ``tree_file``. Note that if
#             either ``aln_file`` or ``tree_file`` are provided, the root must be provided
#             as the sequence name, not as a ``Sequence`` object (as the root sequence must
#             already be included in either ``aln_file`` or ``tree_file``. If the root is not
#             provided, the germline V-gene sequence of the

#         color_dict (dict): Dictionary with sequence IDs as keys and colors (hex format) as values. If any
#             sequence IDs are not found in the dict, they will be colored black. If neither ``color_dict`` nor
#             ``color_function`` is provided, all leaves will be colored black.

#         color_function (func): Function that that accepts a ``Sequence`` object and returns the color
#             (as a hex code). If ``color_dict`` is also provided, ``color_function`` is ignored. Additionally,
#             ``color_function`` will only be used if ``sequences`` are provided. If ``sequences`` are not provided
#             (instead using ``aln_file` or ``tree_file``), ``color_dict`` must be used instead of ``color_function``.

#         orders: a dictionary with sequence IDs as keys and orders (integers) as values.
#             If not provided, only the leaf branches will be colored (if <colors> or
#             <color_function> is provided).

#         chain: build a phylogeny using the given chain ('heavy' or 'light').
#             Default is 'heavy'.

#         filter_function: function used to filter sequences (identity-based clustering, for
#             example). The function should accept a list of Sequence objects and return
#             a list of Sequence objects.

#         just_pairs: if True, compute the phylogeny using only paired sequences.
#             Default (False) will use all sequences of the appropriate chain, paired or not.

#         scale (float): passed to ete3.TreeStyle() to set the scale of the tree figure. Increased
#             scale results in a wider tree.

#         branch_vert_margin (int): passed to ete3.TreeStyle() to set the branch_vertical_margin of
#             the tree figure. Increased branch_vert_margin results in a taller tree.

#         fontsize: size of the leaf labels. Default is 12.

#         show_names: show names of leaf nodes. Options are True (show labels for all leaf nodes),
#             False (don't show labels for any leaf nodes) or a list of sequence IDs for which
#             labels should be shown. Default is True.

#         mirror: flip the orientation of the tree. Default is to draw the tree from left to right.
#             Setting mirror to True results in the tree being drawn from right to left.

#         min_order_fraction: minimum fraction of downstream leaves requried to color a branch.
#             When coloring non-leaf nodes, the earliest 'order' with at least <min_order_fraction>
#             leaf nodes is used. Default is 0.1 (which corresponds to 10%).

#         figname_prefix: by default, figures will be named <lineage_id>.pdf. If prefix='prefix_' and
#             the lineage ID is 'ABC123', the figure file will be named 'prefix_ABC123.pdf'.

#         figname_suffix: by default, figures will be named <lineage_id>.pdf. If suffix='_suffix' and
#             the lineage ID is 'ABC123', the figure file will be named 'ABC123_suffix.pdf'.
#     """

#     from abstar.core.germline import get_imgt_germlines

#     if project_dir is None:
#         print("\nERROR: project_dir is required\n")
#         sys.exit(1)
#     else:
#         project_dir = os.path.abspath(project_dir)

#     # make a name if one isn't provided
#     if name is None:
#         name = "".join(
#             random.choice(string.ascii_uppercase + string.digits) for _ in range(8)
#         )

#     # if sequences are provided, need to process them
#     if sequences is not None and all([arg is None for arg in [aln_file, tree_file]]):
#         sequences = deepcopy(sequences)
#         root = copy(root)

#         # if custom seq_field is specified, copy to the .sequence attribute
#         if seq_field is not None:
#             if not all([seq_field in list(s.annotations.keys()) for s in sequences]):
#                 print(
#                     "\nERROR: {} is not present in all of the supplied sequences.\n".format(
#                         seq_field
#                     )
#                 )
#                 sys.exit(1)
#             for s in sequences:
#                 s.alignment_sequence = s[seq_field]
#         else:
#             for s in sequences:
#                 s.alignment_sequence = s.sequence

#         # if custom name_field is specified, copy to the .id attribute
#         if name_field is not None:
#             if not all([name_field in list(s.annotations.keys()) for s in sequences]):
#                 print(
#                     "\nERROR: {} is not present in all of the supplied sequences.\n".format(
#                         name_field
#                     )
#                 )
#                 sys.exit(1)
#             for s in sequences:
#                 s.alignment_id = s[name_field]
#         else:
#             for s in sequences:
#                 s.alignment_id = s.id

#         # parse the root sequence
#         if unrooted:
#             root = None
#             root_name = None
#         elif root is None:
#             if not quiet:
#                 print(
#                     "\nRoot sequence was was not provided. Using the germline V-gene."
#                 )
#             if not all(["v_gene" in list(s.annotations.keys()) for s in sequences]):
#                 print(
#                     "\nInput sequences to not appear to be AbStar annotated. Annotating now..."
#                 )
#                 sequences = abstar.run(*[(s.id, s.sequence) for s in sequences])
#                 print("Done.")
#             if not all(["full" in list(s["v_gene"].keys()) for s in sequences]):
#                 print(
#                     "\nInput sequences to not appear to be AbStar annotated. Annotating now..."
#                 )
#                 sequences = abstar.run(*[(s.id, s.sequence) for s in sequences])
#                 print("Done.")
#             top_vgene = sorted(
#                 list(Counter([s["v_gene"]["full"] for s in sequences]).items()),
#                 key=lambda x: x[1],
#                 reverse=True,
#             )[0][0]
#             vgene = get_imgt_germlines(species, "V", gene=top_vgene)
#             if aa:
#                 root = Sequence(vgene.ungapped_aa_sequence, id=top_vgene)
#             else:
#                 root = Sequence(vgene.ungapped_nt_sequence, id=top_vgene)
#             root.alignment_id = root.id
#             root.alignment_sequence = root.sequence
#             if not quiet:
#                 print("Top V-gene: {}".format(root.alignment_id))
#                 print(root.alignment_sequence)
#         elif type(root) in STR_TYPES:
#             root = [s for s in sequences if s.alignment_id == root][0]
#             if not root:
#                 print(
#                     "\nERROR: The name of the root sequence ({}) was not found in the list of input sequences.".format(
#                         root
#                     )
#                 )
#                 print("\n")
#                 sys.exit(1)
#             sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
#         elif type(root) == Sequence:
#             if seq_field is not None:
#                 if seq_field not in list(root.anotations.keys()):
#                     print(
#                         "\nERROR: {} is not present in the supplied root sequence.\n".format(
#                             seq_field
#                         )
#                     )
#                     sys.exit(1)
#                 root.alignment_sequence = root[seq_field]
#             if name_field is not None:
#                 if name_field not in list(root.anotations.keys()):
#                     print(
#                         "\nERROR: {} is not present in the supplied root sequence.\n".format(
#                             name_field
#                         )
#                     )
#                     sys.exit(1)
#                 root.alignment_id = root[name_field]
#             sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
#         else:
#             print(
#                 "\nERROR: If root is provided, it must be the name of a sequence \
#             found in the supplied list of sequences or it must be a Sequence object."
#             )
#             print("\n")
#             sys.exit(1)
#         if not unrooted:
#             if root_name is not None:
#                 root.alignment_id = root_name
#             else:
#                 root_name = root.alignment_id
#             sequences.append(root)

#     # parse sequences from aln_file, if provided
#     elif aln_file is not None:
#         if not unrooted and type(root) not in STR_TYPES:
#             print(
#                 "\nERROR: If providing an aln_file, the name of the root sequence must \
#             be provided (as a string) using the root keyword argument"
#             )
#             print("\n")
#             sys.exit(1)
#         _sequences = []
#         _root = None
#         for rec in AlignIO.read(open(aln_file), "fasta"):
#             s = str(rec.seq).replace("-", "")
#             if rec.id == root:
#                 _root = Sequence(s, rec.id)
#                 _root.alignment_id = _root.id
#             else:
#                 _s = Sequence(s, id=rec.id)
#                 _s.alignment_id = rec.id
#                 _sequences.append(_s)
#         if sequences is None:
#             sequences = _sequences
#         else:
#             sequence_ids = [s.id for s in sequences]
#             if any([_s.alignment_id not in sequence_ids for _s in _sequences]):
#                 print(
#                     "\nWARNING: Sequences were found in the alignment file that were not included \
#                       in the input sequence list. This may cause problems."
#                 )
#             for s in sequences:
#                 s.alignment_id = s.id
#                 s.alignment_sequence = s.sequence
#         if unrooted:
#             root = None
#             root_name = None
#         else:
#             if _root is None:
#                 print(
#                     "\nERROR: The specified root ({}) was not found in the provided alignment file.".format(
#                         root
#                     )
#                 )
#                 print("\n")
#                 sys.exit(1)
#             root = _root
#             if root_name is not None:
#                 root.alignment_id = root_name
#             else:
#                 root_name = root.alignment_id
#             sequences = [
#                 s
#                 for s in sequences
#                 if all(
#                     [s.alignment_id != name for name in [root.id, root.alignment_id]]
#                 )
#             ]
#             sequences.append(root)

#     # parse sequences from tree_file, if provided
#     elif tree_file is not None:
#         if not unrooted and type(root) not in STR_TYPES:
#             print(
#                 "\nERROR: If providing a tree_file, the name of the root sequence must \
#             be provided (as a string) using the root keyword argument"
#             )
#             print("\n")
#             sys.exit(1)
#         _sequences = []
#         _root = None
#         tree = Phylo.read(open(tree_file), "newick")
#         for leaf in tree.get_terminals():
#             s = ""
#             if leaf.name == root:
#                 _root = Sequence(s, leaf.name)
#                 _root.alignment_id = _root.id
#             else:
#                 _s = Sequence(s, id=leaf.name)
#                 _s.alignment_id = leaf.name
#                 _sequences.append(_s)
#         if sequences is None:
#             sequences = _sequences
#         else:
#             sequence_ids = [s.id for s in sequences]
#             if any([_s.alignment_id not in sequence_ids for _s in _sequences]):
#                 print(
#                     "\nWARNING: Sequences were found in the alignment file that were not included \
#                       in the input sequence list. This may cause problems."
#                 )
#             for s in sequences:
#                 s.alignment_id = s.id
#                 s.alignment_sequence = s.sequence
#         if unrooted:
#             root = None
#             root_name = None
#         elif _root is None:
#             print(
#                 "\nERROR: The specified root ({}) was not found in the provided tree file.".format(
#                     root
#                 )
#             )
#             print("\n")
#             sys.exit(1)
#         else:
#             root = _root
#             if root_name is not None:
#                 root.alignment_id = root_name
#             else:
#                 root_name = root.alignment_id
#             sequences = [
#                 s
#                 for s in sequences
#                 if all(
#                     [s.alignment_id != name for name in [root.id, root.alignment_id]]
#                 )
#             ]
#             sequences.append(root)

#     # set up colors and color ordering
#     if order_dict is None:
#         if order_function is not None:
#             order_dict = {seq.alignment_id: order_function(seq) for seq in sequences}
#     if color_dict is None:
#         if color_function is not None:
#             color_dict = {seq.alignment_id: color_function(seq) for seq in sequences}
#     if color_dict is None:
#         color_dict = {}

#     # make msa (if necessary)
#     if all([aln_file is None, tree_file is None]):
#         aln_file = os.path.abspath(os.path.join(project_dir, "{}.aln".format(name)))
#         # muscle(seqs, aln_file, as_file=True)
#         do_print = False if quiet else True
#         if do_print:
#             print("\n")
#         seqs = [(s.alignment_id, s.alignment_sequence) for s in sequences]
#         mafft(
#             seqs, aln_file, as_file=True, print_stdout=do_print, print_stderr=do_print
#         )

#     # make treefile (if necessary)
#     if tree_file is None:
#         tree_file = os.path.abspath(os.path.join(project_dir, "{}.nw".format(name)))
#         fasttree(aln_file, tree_file, is_aa=aa, quiet=quiet)

#     # make phylogeny
#     prefix = "" if figname_prefix is None else figname_prefix
#     suffix = "" if figname_suffix is None else figname_suffix
#     fig_file = os.path.join(project_dir, "{}{}{}.pdf".format(prefix, name, suffix))
#     _make_tree_figure(
#         tree_file,
#         fig_file,
#         color_dict,
#         order_dict,
#         None if root is None else root.alignment_id,
#         rename_function=rename_function,
#         show_names=show_names,
#         name_field=name_field,
#         branch_vert_margin=branch_vert_margin,
#         scale=scale,
#         color_node_labels=color_node_labels,
#         label_colors=label_colors,
#         show_root_name=show_root_name,
#         tree_orientation=1 if mirror else 0,
#         fontsize=fontsize,
#         min_order_fraction=min_order_fraction,
#         #   linked_alignment=linked_alignment,
#         #   alignment_fontsize=alignment_fontsize,
#         #   alignment_height=alignment_height,
#         #   alignment_width=alignment_width,
#         show_scale=show_scale,
#         compact_alignment=compact_alignment,
#         scale_factor=scale_factor,
#         linewidth=linewidth,
#         ladderize=ladderize,
#         delete_nodes=delete_nodes,
#     )


# def _make_tree_figure(
#     tree,
#     fig,
#     colors,
#     orders,
#     root_name,
#     scale=None,
#     branch_vert_margin=None,
#     fontsize=12,
#     show_names=True,
#     name_field="seq_id",
#     rename_function=None,
#     color_node_labels=False,
#     label_colors=None,
#     tree_orientation=0,
#     min_order_fraction=0.1,
#     show_root_name=False,
#     chain=None,
#     # linked_alignment=None, alignment_fontsize=11, alignment_height=50, alignment_width=50,
#     compact_alignment=False,
#     scale_factor=1,
#     linewidth=1,
#     show_scale=False,
#     ladderize=True,
#     delete_nodes=None,
# ):
#     if delete_nodes is None:
#         delete_nodes = []
#     elif type(delete_nodes) in STR_TYPES:
#         delete_nodes = [
#             delete_nodes,
#         ]
#     if show_root_name is True:
#         show_names.append(root_name)
#     # if linked_alignment is not None:
#     #     t = ete3.PhyloTree(tree, alignment=linked_alignment, alg_format='fasta')
#     #     ete3.faces.SequenceItem = MySequenceItem
#     t = ete3.Tree(tree)
#     if root_name is not None:
#         t.set_outgroup(t & root_name)
#     # style the nodes
#     for node in t.traverse():
#         if node.name in delete_nodes:
#             node.delete()
#             continue
#         if orders is not None:
#             leaves = node.get_leaf_names()
#             order_count = Counter([orders[l] for l in leaves])
#             for order in sorted(order_count.keys()):
#                 if float(order_count[order]) / len(leaves) >= min_order_fraction:
#                     color = colors[order]
#                     break
#         else:
#             color = colors.get(node.name, "#000000")
#         # if linked_alignment is not None:
#         #     node.add_feature('aln_fontsize', alignment_fontsize)
#         #     node.add_feature('aln_height', alignment_height)
#         #     node.add_feature('aln_width', alignment_width)
#         #     node.add_feature('fontsize', fontsize)
#         #     node.add_feature('format', 'seq')
#         #     node.add_feature('scale_factor', scale_factor)
#         style = ete3.NodeStyle()
#         style["size"] = 0
#         style["vt_line_width"] = float(linewidth)
#         style["hz_line_width"] = float(linewidth)
#         style["vt_line_color"] = color
#         style["hz_line_color"] = color
#         style["vt_line_type"] = 0
#         style["hz_line_type"] = 0
#         if show_names is True:
#             tf = _build_node_text_face(
#                 node, color_node_labels, color, label_colors, fontsize, rename_function
#             )
#             node.add_face(tf, column=0)
#         elif node.name in show_names:
#             tf = _build_node_text_face(
#                 node, color_node_labels, color, label_colors, fontsize, rename_function
#             )
#             node.add_face(tf, column=0)
#         node.set_style(style)
#     t.dist = 0
#     ts = ete3.TreeStyle()
#     # if linked_alignment is not None:
#     #     ts.layout_fn = _phyloalignment_layout_function
#     ts.orientation = tree_orientation
#     ts.show_leaf_name = False
#     if scale is not None:
#         ts.scale = int(scale)
#     if branch_vert_margin is not None:
#         ts.branch_vertical_margin = float(branch_vert_margin)
#     ts.show_scale = show_scale
#     if ladderize:
#         t.ladderize()
#     t.render(fig, tree_style=ts)


# def _build_node_text_face(
#     node, color_node_labels, color, label_colors, fontsize, rename_function
# ):
#     if color_node_labels:
#         if label_colors is None:
#             node_color = color
#         elif type(label_colors) == dict:
#             node_color = label_colors.get(node.name, "#000000")
#         elif type(label_colors) in [list, tuple]:
#             node_color = color if node.name in label_colors else "#000000"
#         else:
#             node_color = "#000000"
#     else:
#         node_color = "#000000"
#     node_name = node.name if rename_function is None else rename_function(node.name)
#     tf = ete3.TextFace(node_name, fsize=fontsize, fgcolor=node_color)
#     return tf


# # -----------------------------------------
# #      PHYLOGENETIC LAPLACIAN SPECTRA
# # -----------------------------------------


# class PLS:
#     """
#     Object representing a collection of Phylogenetic Laplacian Spectra.
#     """

#     def __init__(self, spectral_trees, n_processes=0):
#         self.trees = spectral_trees
#         self.n_processes = mp.cpu_count() if n_processes == 0 else n_processes
#         self.compute_kdes()

#     @lazy_property
#     def distance_matrix(self):
#         dist = {}
#         p = mp.Pool(processes=self.n_processes)
#         async_results = []
#         for t in self.trees:
#             async_results.append(
#                 p.apply_async(self.calc_distance_row, args=(t, self.trees))
#             )
#         dists = [ar.get() for ar in async_results]
#         p.close()
#         p.join()
#         for t1, row in zip(self.trees, dists):
#             for t2, d in zip(self.trees, row):
#                 dist[t1.name][t2.name] = d
#         return pd.DataFrame(dist)

#         # for t1, t2 in itertools.product(self.trees, repeat=2):
#         #     dist[t1.name] = {}
#         #     if t1.name not in dist:
#         #         dist[t1.name] = {}
#         #     dist[t1.name][t2.name] = self.compare_trees(t1, t2)
#         # return pd.DataFrame(dist)

#     def calc_distance_row(self, t1, trees):
#         dists = []
#         for t2 in trees:
#             dists.append(self.compare_trees(t1, t2))
#         return dists

#     def compare_trees(self, t1, t2, n_points=100):
#         maxw = max([t1.principal_eigenvalue, t2.principal_eigenvalue])
#         x = np.linspace(0, maxw, n_points)
#         p = t1.pdf(x)
#         q = t2.pdf(x)
#         return self.jensenshannon(p, q)

#     @staticmethod
#     def jensenshannon(p, q, base=None):
#         p = np.asarray(p)
#         q = np.asarray(q)
#         p = p / np.sum(p, axis=0)
#         q = q / np.sum(q, axis=0)
#         m = (p + q) / 2.0
#         left = rel_entr(p, m)
#         right = rel_entr(q, m)
#         js = np.sum(left, axis=0) + np.sum(right, axis=0)
#         if base is not None:
#             js /= np.log(base)
#         return np.sqrt(js / 2.0)

#     def compute_kdes(self):
#         if self.n_processes == 1:
#             self._compute_kdes_sp()
#         else:
#             self._compute_kdes_mp()

#     def _compute_kdes_sp(self):
#         for t in self.trees:
#             kde = t.kde

#     def _compute_kdes_mp(self):
#         p = mp.Pool(processes=self.n_processes)
#         async_results = []
#         for t in self.trees:
#             async_results.append(p.apply_async(self.compute_kde, args=(t,)))
#         kdes = [ar.get() for ar in async_results]
#         for t, kde in zip(self.trees, kdes):
#             t.kde = kde
#         p.close()
#         p.join()

#     @staticmethod
#     def compute_kde(tree):
#         """
#         For a SpectralTree, computes the KDE.

#         Args:

#             tree (SpectralTree): tree for which the KDE will be computed

#         Returns:

#             the KDE function
#         """
#         return tree.kde

#     def clustermap(self, cmap=None, figfile=None):
#         cm = sns.clustermap(self.distance_matrix, cmap=cmap)
#         if figfile is not None:
#             cm.savefig(figfile)
#         else:
#             plt.show()


# class SpectralTree:
#     def __init__(self, tree_file, subject=None):
#         self.subject = subject
#         self.name = os.path.basename(tree_file)
#         self.tree_file = tree_file
#         self.t = ete3.Tree(tree_file)

#     @lazy_property
#     def nodes(self):
#         num = 0
#         nodes = []
#         for n in self.t.traverse():
#             if not n.name:
#                 n.name = str(num)
#                 num += 1
#             nodes.append(n)
#         return nodes

#     @lazy_property
#     def node_names(self):
#         return [n.name for n in self.nodes]

#     @lazy_property
#     def distances(self):
#         distances = []
#         for n1 in self.nodes:
#             dist = {}
#             for n2 in self.nodes:
#                 if n1 == n2:
#                     continue
#                 dist[n2] = self.t.get_distance(n1, n2)
#             dist[n1] = -1.0 * sum(dist.values())
#             distances.append([dist[n] for n in self.nodes])
#         return np.asarray(distances)

#     @lazy_property
#     def eigenvalues(self):
#         gl = scipy.sparse.csgraph.laplacian(self.distances)
#         w, v = np.linalg.eig(gl)
#         return [e for e in sorted(w, reverse=True) if e >= 1]

#     @lazy_property
#     def log_eigenvalues(self):
#         return np.real(np.log(self.eigenvalues))

#     @property
#     def principal_eigenvalue(self):
#         return max(self.eigenvalues)

#     @lazy_property
#     def kde(self):
#         return scipy.stats.gaussian_kde(self.eigenvalues)

#     def pdf(self, x):
#         return [v.real for v in self.kde.pdf(x)]

#     def plot(self, xs=None, figfile=None, xlim=None):
#         if xs is None:
#             xs = np.linspace(-0.5, self.principal_eigenvalue, 200)
#         ys = self.pdf(xs)
#         # initialize and plot
#         plt.figure(figsize=[3, 4])
#         plt.plot(xs, ys, lw=2)
#         plt.fill_between(xs, ys, alpha=0.2)
#         # style
#         ax = plt.gca()
#         ax.set_xlabel("eigenvalue (ln)")
#         ax.set_ylabel("density")
#         if xlim is not None:
#             ax.set_xlim(xlim)
#         if figfile is not None:
#             plt.tight_layout()
#             plt.savefig(figfile)
#         else:
#             plt.show()



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

from typing import Optional

import click


def parse_dict_from_string(
    ctx: click.Context, param: click.Parameter, value: Optional[str]
) -> dict:
    """
    Parse a string of key-value pairs into a dictionary. Used as a ``click`` callback function
    for parsing command-line arguments.

    Parameters
    ----------
    ctx : click.Context
        The click context.

    param : click.Parameter
        The click parameter.

    value : Optional[str]
        The value to parse. If ``None``, return an empty dictionary.

    Returns
    -------
    dict
        The parsed dictionary.

    Raises
    ------
    click.BadParameter
        If the format is not similar to ``key1=val1,key2=val2``.

    """
    if not value:
        return {}
    try:
        # split the string into key-value pairs
        kv_pairs = value.split(",")
        # convert the pairs into a dictionary
        return dict(kv_pair.split("=") for kv_pair in kv_pairs)
    except ValueError:
        raise click.BadParameter("Format must be 'key1=val1,key2=val2'")


class HiddenClickOption(click.Option):
    """
    Custom ``click`` option that is hidden from the CLI help message.
    """

    def get_help_record(self, ctx):
        return None



#!/usr/bin/env python
# filename: heatmap.py


#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import sys
from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd

import matplotlib as mpl
import matplotlib.pyplot as plt

from .models.data import PlotData
from ..utils.color import get_cmap


def heatmap(
    data: Union[pd.DataFrame, Iterable],
    color: Union[str, Iterable, mpl.colors.Colormap] = "Greys",
    row_colors: Optional[Iterable] = None,
    column_colors: Optional[Iterable] = None,
    scale_color_by_row: bool = False,
    scale_color_by_column: bool = False,
    transform: Optional[str] = None,
    norm: bool = False,
    percent: bool = False,
    norm_axis: Union[int, str] = "columns",
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    row_order: Optional[Iterable] = None,
    row_labels: Optional[Iterable] = None,
    row_label_position: str = "left",
    row_label_rotation: float = 0.0,
    row_label_fontsize: float = 14.0,
    show_row_ticks: bool = False,
    column_order: Optional[Iterable] = None,
    column_labels: Optional[Iterable] = None,
    column_label_position: str = "top",
    column_label_rotation: float = 90.0,
    column_label_fontsize: float = 14.0,
    show_column_ticks: bool = False,
    show_values: bool = False,
    values_fmt: Union[str, mpl.ticker.Formatter] = "{x:0.1f}",
    values_fontsize: float = 9.0,
    values_fontweight: str = "normal",
    values_color: Union[str, Iterable, None] = None,
    values_lightcolor: Union[str, Iterable] = "#FFFFFF",
    values_darkcolor: Union[str, Iterable] = "#303030",
    values_horizontalalignment: str = "center",
    values_verticalalignment: str = "center",
    values_color_threshold: Optional[float] = None,
    square: bool = True,
    linewidth: float = 2.0,
    linecolor: str = "#FFFFFF",
    linestyle: str = "-",
    ax: Optional[mpl.axes.Axes] = None,
    figsize: Optional[Iterable] = None,
    figfile: Optional[str] = None,
) -> Optional[mpl.axes.Axes]:
    """
    Creates a heatmap from a pandas DataFrame or an iterable of iterables.

    Parameters
    ----------
    data : Union[pd.DataFrame, Iterable]
        A pandas DataFrame or an iterable of iterables. If a DataFrame is provided,
        the index and columns will be used as the row and column labels, respectively.

    color : Union[str, Iterable, mpl.colors.Colormap], optional
        The color palette to use for the heatmap. Can be any of the following:
          * a ``list`` or ``tuple`` containing RGB or RGBA values
          * a color string, either from `Matplotlib's set of named colors`_ or a hex color value
          * a Matplotlib ``Colormap`` object
          * the name of a registered `Matplotlib Colormap`_
        Default is ``'Greys'``.

    row_colors : Optional[Iterable], optional
        A list of colors to use for each row. The number of colors must match the number of rows.
        Each list item can be anything accepted by `color`. If `row_colors` is provided,
        `color` will be ignored. Default is ``None``.

    column_colors : Optional[Iterable], optional
        A list of colors to use for each column. The number of colors must match the number of
        columns. Each list item can be anything accepted by `color`. If `column_colors` is
        provided, `color` will be ignored. Default is ``None``.

    scale_color_by_row : bool, optional
        If ``True``, the color values will be separately scaled for each row.
        If both `scale_color_by_row` and `scale_color_by_column` are ``False``, the color values
        will be scaled across the entire heatmap. Default is ``False``.

    scale_color_by_column : bool, optional
        If ``True``, the color values will be separately scaled for each column.
        If both `scale_color_by_row` and `scale_color_by_column` are ``False``, the color values
        will be scaled across the entire heatmap. Default is ``False``.

    transform : Optional[str], optional
        The type of transformation to apply to the data. Can be any of the following:
            * ``'log'``: apply a log10 transformation
            * ``'log2'``: apply a log2 transformation
            * ``'log10'``: apply a log10 transformation
            * ``'linear'``: do not apply a transformation
        Default is ``None``, which has the same effect as ``'linear'``.

    vmin : Optional[float], optional
        The minimum value to use for the color scale. If ``None``, the minimum value in the data
        will be used. Default is ``None``.

    vmax : Optional[float], optional
        The maximum value to use for the color scale. If ``None``, the maximum value in the data
        will be used. Default is ``None``.

    row_order : Optional[Iterable], optional
        A list of row labels in the order they should appear in the heatmap. If ``None``, the
        rows will be ordered by the index of the DataFrame. Default is ``None``.

    row_labels : Optional[Iterable], optional
        A list of labels to use for each row. If ``None``, the row labels will be the index of the
        DataFrame. Default is ``None``.

    row_label_position : str, optional
        The position of the row labels. Can be either ``'left'`` or ``'right'``. Default is
        ``'left'``.

    row_label_rotation : float, optional
        The rotation of the row labels in degrees. Default is ``0.0``.

    row_label_fontsize : float, optional
        The font size of the row labels. Default is ``14``.

    show_row_ticks : bool, optional
        If ``True``, ticks will be shown with the row labels. Default is ``False``.

    column_order : Optional[Iterable], optional
        A list of column labels in the order they should appear in the heatmap. If ``None``, the
        columns will be ordered by the columns of the DataFrame. Default is ``None``.

    column_labels : Optional[Iterable], optional
        A list of labels to use for each column. If ``None``, the column labels will be the columns
        of the DataFrame. Default is ``None``.

    column_label_position : str, optional
        The position of the column labels. Can be either ``'top'`` or ``'bottom'``. Default is
        ``'top'``.

    column_label_rotation : float, optional
        The rotation of the column labels in degrees. Default is ``0.0``.

    column_label_fontsize : float, optional
        The font size of the column labels. Default is ``14``.

    show_column_ticks : bool, optional
        If ``True``, ticks will be shown with the column labels. Default is ``False``.

    show_values : bool, optional
        If ``True``, the values will be shown in each cell. Default is ``False``.

    values_fmt : str, optional
        The format string to use for the values. Default is ``'{x:0.2f}'``.

    values_fontsize : float, optional
        The font size of the values. Default is ``8``.

    values_fontweight : str, optional
        The font weight of the values. Default is ``'normal'``.

    values_color : Union[str, Iterable], optional
        The color of the values. Can be a color string, either from
        `Matplotlib's set of named colors`_ or a hex color value, or a list of RGB(A)
        values. Default is ``None``, which will use `values_lightcolor`
        for dark backgrounds and `values_darkcolor` for light backgrounds.

    values_lightcolor : Union[str, Iterable]
        The color of the values when the background is dark. Can be a color string, either from
        `Matplotlib's set of named colors`_ or a hex color value, or a list of RGB(A) values.
        Default is ``'#FFFFFF'`` (white).

    values_darkcolor : Union[str, Iterable]
        The color of the values when the background is light. Can be a color string, either from
        `Matplotlib's set of named colors`_ or a hex color value, or a list of RGB(A) values.
        Default is ``'#303030'`` (dark gray).

    values_horizontalalignment : str, optional
        The horizontal alignment of the values. Can be either ``'left'``, ``'center'``, or
        ``'right'``. Default is ``'center'``.

    values_verticalalignment : str, optional
        The vertical alignment of the values. Can be either ``'top'``, ``'center'``, or
        ``'bottom'``. Default is ``'center'``.

    values_color_threshold : Optional[float], optional
        The threshold to use for determining whether to use `values_lightcolor` or
        `values_darkcolor`. If ``None``, the threshold will be set to the midpoint of the
        background color scale. Default is ``None``.

    square : bool, optional
        If ``True``, heatmap boxes will be square. Default is ``True``.

    linewidth : float, optional
        The width of the lines that will divide each box. Default is ``2.0``.

    linecolor : Union[str, Iterable], optional
        The color of the lines that will divide each box. Can be a color string, either from
        `Matplotlib's set of named colors`_ or a hex color value, or a list of RGB(A) values.
        Default is ``'#FFFFFF'`` (white).

    linestyle : str, optional
        The style of the lines that will divide each box. Default is ``'-'``, which is a
        solid line.

    ax : Optional[mpl.axes.Axes], optional
        The `Matplotlib Axes`_ on which to draw the heatmap. If ``None``, a new figure and axes
        will be created. Default is ``None``.

    figsize : Optional[Tuple[float, float]], optional
        The width and height of the figure in inches. Default is ``None``, which will use
        the dimensions of the input data to infer an appropriate size.

    figfile : Optional[str], optional
        The path to save the figure to. If ``None``, the figure will not be saved and
        the figure ``Axes`` will be returned. Default is ``None``.

    Returns
    -------
    Optional[mpl.axes.Axes]
        The `Matplotlib Axes`_ on which the heatmap was drawn. If ``figfile`` is not ``None``,
        the figure will be saved and the ``Axes`` will be returned.


    .. _Matplotlib's set of named colors:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    -- _Matplotlib Colormap:
        https://matplotlib.org/stable/tutorials/colors/colormaps.html
    .. _Matplotlib Axes:
        https://matplotlib.org/stable/api/axes_api.html

    """
    data = PlotData(data=data, column_labels=column_labels, row_labels=row_labels,)

    # reorder data
    data.reorder(row_order=row_order, column_order=column_order)
    column_labels = data.df.columns.values.tolist()
    row_labels = data.df.index.values.tolist()

    # norm
    if norm or percent:
        data.norm(axis=norm_axis, as_percent=percent)

    # freeze values
    val_df = data.df.copy()
    val_data = val_df.to_numpy()
    n_rows, n_columns = val_df.shape

    # clip data
    data.clip(lower=vmin, upper=vmax)

    # transform
    data.transform(func=transform)

    # scale
    if scale_color_by_row:
        data.scale(axis="rows")
    elif scale_color_by_column:
        data.scale(axis="columns")
    else:
        data.scale()

    # colors
    if row_colors is not None:
        if (n_row_colors := len(row_colors)) < n_rows:
            err = f"\nERROR: the number of row_colors must be equal or greater to the number of rows in the input dataset.\n"
            err += f"The input dataset has {n_rows} rows, but {n_row_colors} colors were provided.\n"
            print(err)
            sys.exit()
        row_cmaps = [get_cmap(c) for c in row_colors]
        color_df = get_color_values(data.df, cmap=row_cmaps, by_row=True)
    elif column_colors is not None:
        if (n_column_colors := len(column_colors)) < n_columns:
            err = f"\nERROR: the number of row_colors must be equal or greater to the number of rows in the input dataset.\n"
            err += f"The input dataset has {n_columns} rows, but {n_column_colors} colors were provided.\n"
            print(err)
            sys.exit()
        column_cmaps = [get_cmap(c) for c in column_colors]
        color_df = get_color_values(data.df, cmap=column_cmaps, by_column=True)
    else:
        cmap = get_cmap(color)
        color_df = get_color_values(data.df, cmap=cmap)

    # reformat the color data
    #
    # color_df.to_numpy() returns a 2D array with tuples (of type "object")
    # as values, rather than a 3D Numpy array. Calling .tolist() on the
    # 2D array retypes the values and allows me to create a 3D
    # array with np.array()
    color_data = np.array(color_df.to_numpy().tolist())

    # set figsize and aspect
    if figsize is None:
        figsize = np.array(color_df.shape[::-1]) / 2
    plt.figure(figsize=figsize)

    # plot
    im = plt.imshow(color_data, aspect="equal" if square else "auto")
    ax = plt.gca()

    # row label positions
    if row_label_position.lower() == "right":
        labelleft = False
        labelright = True
        tickleft = False
        tickright = True if show_row_ticks else False
    else:
        labelleft = True
        labelright = False
        tickleft = True if show_row_ticks else False
        tickright = False
    # row labels
    if row_labels is None:
        row_labels = val_df.index.values
    ax.set_yticks(np.arange(val_df.shape[0]), labels=row_labels)
    ax.tick_params(
        axis="y", labelrotation=row_label_rotation, labelsize=row_label_fontsize,
    )

    # column label positions
    if column_label_position.lower() == "top":
        labeltop = True
        labelbottom = False
        ticktop = True if show_column_ticks else False
        tickbottom = False
    else:
        labeltop = False
        labelbottom = True
        ticktop = False
        tickbottom = True if show_column_ticks else False
    # column labels
    if column_labels is None:
        column_labels = val_df.columns.values
    ax.set_xticks(np.arange(val_df.shape[1]), labels=column_labels)
    ax.tick_params(
        axis="x", labelrotation=column_label_rotation, labelsize=column_label_fontsize,
    )

    # show/hide ticks and labels
    ax.tick_params(
        top=ticktop,
        bottom=tickbottom,
        right=tickright,
        left=tickleft,
        labeltop=labeltop,
        labelbottom=labelbottom,
        labelleft=labelleft,
        labelright=labelright,
    )

    # hide the spines
    ax.spines[:].set_visible(False)

    # draw gridlines separating the heatmap boxes
    ax.set_xticks(np.arange(val_df.shape[1] + 1) - 0.5, minor=True)
    ax.set_yticks(np.arange(val_df.shape[0] + 1) - 0.5, minor=True)
    ax.grid(
        which="minor", color=linecolor, linestyle=linestyle, linewidth=linewidth,
    )
    ax.tick_params(which="minor", bottom=False, left=False, top=False, right=False)

    # show values
    if show_values:
        # set the threshold for light/dark text
        if values_color_threshold is not None:
            threshold = values_color_threshold
        else:
            threshold = val_data.max() / 2

        # get the string formatter for the values
        if isinstance(values_fmt, str):
            values_fmt = mpl.ticker.StrMethodFormatter(values_fmt)

        # write the values
        lightcolor = values_color if values_color is not None else values_lightcolor
        darkcolor = values_color if values_color is not None else values_darkcolor
        for i in range(val_data.shape[0]):
            for j in range(val_data.shape[1]):
                c = lightcolor if val_data[i, j] > threshold else darkcolor
                ax.text(
                    x=j,
                    y=i,
                    s=values_fmt(val_data[i, j], None),
                    color=c,
                    horizontalalignment=values_horizontalalignment,
                    verticalalignment=values_verticalalignment,
                    fontsize=values_fontsize,
                    fontweight=values_fontweight,
                )

    # save or return
    if figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    else:
        return ax


def get_color_values(
    df: pd.DataFrame,
    cmap: Union[mpl.colors.Colormap, Iterable[mpl.colors.Colormap]],
    by_row=False,
    by_column=False,
) -> pd.DataFrame:
    """
    Gets colors for each value in a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing values to be colored.

    cmap : Union[mpl.colors.Colormap, Iterable[mpl.colors.Colormap]]
        ``Colormap`` or iterable of ``Colormap`` objects.

    by_row : bool, default=False
        Whether to get colors by row.

    by_column : bool, default=False
        Whether to get colors by column.

    Returns
    -------
    color_df : pd.DataFrame
        DataFrame containing colors for each value in ``df``.
    """
    df = df.copy()
    if by_column:
        return get_colors_by_column(df, cmaps=cmap)
    elif by_row:
        return get_colors_by_row(df, cmaps=cmap)
    else:
        for c in df.columns:
            df[c] = [cmap(v) for v in df[c]]
        return df


def get_colors_by_column(
    df: pd.DataFrame, cmaps: Iterable[mpl.colors.Colormap]
) -> pd.DataFrame:
    """
    Gets colors for each value in a DataFrame, by column.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing values to be colored.

    cmaps : Iterable[mpl.colors.Colormap]
        Iterable of ``Colormap`` objects, one for each column in ``df``.

    Returns
    -------
    color_df : pd.DataFrame
        DataFrame containing colors for each value in ``df``.
    """
    for c, cmap in zip(df.columns, cmaps):
        df[c] = [cmap(v) for v in df[c]]
    return df


def get_colors_by_row(
    df: pd.DataFrame, cmaps: Iterable[mpl.colors.Colormap]
) -> pd.DataFrame:
    """
    Gets colors for each value in a DataFrame, by row.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing values to be colored.

    cmaps : Iterable[mpl.colors.Colormap]
        Iterable of ``Colormap`` objects, one for each row in ``df``.

    Returns
    -------
    color_df : pd.DataFrame
        DataFrame containing colors for each value in ``df``.
    """
    df = get_colors_by_column(df.T, cmaps=cmaps)
    return df.T



#!/usr/bin/env python
# filename: donut.py


#
# Copyright (c) 2022 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from collections import Counter
from typing import Union, Iterable, Optional, Callable

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl

import seaborn as sns

from natsort import natsorted, natsort_keygen

from ..utils.color import get_cmap, monochrome_palette


def donut(
    values: Union[Iterable, str],
    counts: Union[Iterable, str, None] = None,
    hue: Union[Iterable, str, None] = None,
    data: Optional[pd.DataFrame] = None,
    palette: Optional[dict] = None,
    color: Union[str, Iterable, None] = None,
    cmap: Optional[mpl.colors.Colormap] = None,
    order: Optional[Iterable] = None,
    hue_order: Optional[Iterable] = None,
    hue_agg: Union[str, Callable] = "most common",
    sort_by: str = "count",
    sort_key: Callable = natsort_keygen(),
    sort_descending: bool = False,
    force_categorical_hue: bool = False,
    force_continuous_hue: bool = False,
    alt_color: Union[str, Iterable] = "#F5F5F5",
    edgecolor: Union[str, Iterable] = "white",
    singleton_color: Union[str, Iterable, None] = None,
    group_singletons: bool = True,
    shuffle_colors: bool = False,
    random_seed: Union[int, float, str] = 1234,
    title: Optional[str] = None,
    title_fontsize: Union[int, float] = 36,
    title_x: Union[int, float] = 0.5,
    title_y: Union[int, float] = 0.5,
    subtitle: Optional[str] = None,
    subtitle_fontsize: Union[int, float] = 20,
    subtitle_x: Union[int, float] = 0.5,
    subtitle_y: Union[int, float] = 0.425,
    show_subtitle: bool = True,
    width: Union[int, float] = 0.55,
    linewidth: Union[int, float] = 2,
    plot_kwargs: Optional[dict] = None,
    text_kwargs: Optional[dict] = None,
    subtext_kwargs: Optional[dict] = None,
    ax: Optional[mpl.axes.Axes] = None,
    show: bool = False,
    figsize: Optional[Iterable] = None,
    figfile: Optional[str] = None,
) -> Optional[mpl.axes.Axes]:
    """
    Creates a donut plot of a population of lineages, with arc widths proportional to lineage size.

    .. note::
       For **continuous** hues (for example, AgBC UMI counts), the mean value for each lineage is used. 
       For **boolean** hues (for example, specificity classifications), the lineage is considered ``True`` if
       any lineage member is ``True``. For **categorical** hues (for example, CDR3 length), the most common
       value for each lineage is used. 
    
    Parameters
    ----------
    values : str or Iterable
        Name of a column in `data` or an iterable of values to be used as segments of 
        the donut plot. Required.

    counts : str or Iterable
        Name of a column in `data` or an iterable of counts used to determine the size
        of each donut segment. If not provided, counts will be computed by calling 
        ``Counter()`` on `values`.
            
    hue : str or dict, optional  
        Can be either the name of a column in `data` or a ``dict`` mapping
        values to hues. If a ``dict`` is provided, any missing values will 
        still be included in the donut plot but will be colored using `alt_color`. 
        There are three possible classes of hues:  
             
                - **continuous:** hues that map to a continuous numerical space. Hues are
                  automatically calssified as continuous if all `hue` values are ``float``s. 
                  If `force_continuous_hue` is ``True``, the hue will be considered 
                  continuous regardless of whether all hue values are ``float``s.   
                  
                - **boolean:** hues that map to either ``True`` or ``False``. For boolean 
                  hues, if the `hue` for any member of a group of values is ``True``, 
                  the entire group of values will be considered ``True``.  
                  
                - **categorical:** hues that map to one of a set of categories. For 
                  categorical hues, the most common hue for a group of values will be 
                  used for the entire group.   
                  
        Finally, if `hue` is not provided, `values` will be considered the `hue`, and
        each unique value will be colored separately.
            
    palette : dict, default=None  
        A ``dict`` mapping `hue` categories to colors. For boolean `hue` types, if `palette`
        is not provided, `color` will be used for ``True`` and `alt_color` will be used for 
        ``False``. For categorical `hue` types, if neither `color` nor `palette` is provided, 
        ``seaborn.hls_palette`` will be used to generate a color palette. For continuous `hue` 
        types, a ``Colormap`` will be used and `palette` is ignored.  
            
    color : str or list, default=None  
        A `Matplotlib color`_ name, hex string or RGB list/tuple for coloring the donut plot. For
        boolean `hue` types, `color` will be used for ``True`` and 'alt_color' will be used for 
        ``False``. For categorical and continuous hue types, a monochromatic palette will 
        be created that contains various shades from `color` to white (but not including white). 
        If not provided, the first color from ``seaborn.color_palette`` is used.  

    cmap : str or matplotlib.color.Colormap, default=None   
        Colormap to be used for continuous `hue` data. If not provided, `color` is used to 
        make a monochromatic ``Colormap``. If `color` is not provided, Matplotlib's built-in
        ``"flare"`` ``Colormap`` is used.

    order : list, default=None  
        A list specifying the order of donut segments. Elements in `order` should be `values` 
        categories. If not provided, segments will be ordered using `sort_by` and `sort_key`.

    hue_order : list, default=None  
        A list specifying the order of hue categories. This does not affect the ordering 
        of segments in the donut plot, just the assignment of colors to `hue` categories. 
        For example, when plotting with a monochromatic palette (by providing `color`), 
        `hue` categories will be colored from dark to light in `hue_order` order. To re-order 
        donut segments by hue categories, use `sort_by`.

    hue_agg : str or callable, default="most common"
        Method by which to aggregate `hue` values. Can be a ``str`` or a function. Built-in
        options are:
            - **most common:** selects the most common hue value. Suitable for all hue types.
            - **max**: selects the maximum value. Suitable primarily for numerical hue types.
            - **min**: selects the minimum value. Suitable primarily for numerical hue types.
            - **mean**: selects the average value. Suitable primarily for numerical hue types.
            - **median**: selects the median value. Suitable primarily for numerical hue types.
        If a callable is provided, it should accept a list of hue values and return a single 
        value that will be used for the entire category.

    sort_by : str, default="count"
        Field to be used to sort the donut segments. Default is ``"count"``, which sorts 
        the segments by size. Options are:
            - **count:** sorts the segments by the `counts` category (which is the segment size)
            - **value:** sorts the segments by the `values` category 
            _ **hue:** sorts the segments by the `hue` category

    sort_key : callable, default=natsort.natsort_keygen()
        Callable function that determines the sorting method. Will be passed directly to 
        ``sorted()`` as the ``key`` argument.  

    sort_descending : bool, default=False
        If ``True``, sort donut segments in descending order.  

    force_categorical_hue : bool, default=False  
        If ``True``, `hue` categories will always be considered categorical. 

    force_continuous_hue : bool, default=False  
        If ``True``, `hue` categories will always be considered continuous. 
            
    alt_color : str or list, default='#F5F5F5'    
        A color name, hex string or RGB list/tuple for coloring alternate values 
        (``False`` boolean hues or values not found in `palette`). Default is 
        ``'#F5F5F5'``, which maps to a very light grey.  

    edgecolor : str or list, default='white'  
        A `Matplotlib color`_ name, hex string or RGB list/tuple for coloring the edges that 
        separate donut segments. 
        
    singleton_color : str or list, default=None  
        A `Matplotlib color`_ name, hex string or RGB list/tuple for coloring the singleton 
        segment of the donut plot. Only used if `group_singletons` is ``True``. If not provided,
        `alt_color` is used.  

    group_singletons : bool, default=False
        Group all singletons (`values` categories with a `count` of ``1``) into a single, 
        undivided segment.
            
    shuffle_colors : bool, default=False  
        If ``True``, colors will be shuffled prior to assignment to hue categories. This
        can be useful when a monochromatic palette is used, in order to make it easier to 
        distinguish neighboring segments on the plot.   
            
    random_seed : int, float or str, default=1234  
        Used to set the random seed using ``numpy.random.seed()``. Only applicable when 
        `shuffle_colors` is ``True``, and provided mainly to allow users to recreate plots
        that use shuffled colors (otherwise the shuffle order would be random, thus creating
        a different color order each time the plotting function is called).  

    title : str, default=None
        Title, printed in the center of the donut "hole". If not provided, the total 
        count of all elements in `values` is used.

    title_fontsize : float or int, default=36
        Fontsize for the title.

    title_x : float, default=0.5
        X-axis position for the center of the `title` text, in relative coordinates. 
        Should be in the range of 0-1. Default is ``0.5``, which is centered horizontally.

    title_y : float, default=0.5
        Y-axis position for the center of the `title` text, in relative coordinates. 
        Should be in the range of 0-1. Default is ``0.5``, which is centered vertically.  

    subtitle : str, default=None
        Title, printed in the center of the donut "hole". If not provided, the total 
        count of all elements in `values` is used, formatted like ``"n={count}"``.

    subtitle_fontsize : float or int, default=20
        Fontsize for the title.

    subtitle_x : float, default=0.5
        X-axis position for the center of the `subtitle` text, in relative coordinates. 
        Should be in the range of 0-1. Default is ``0.5``, which is centered horizontally.

    subtitle_y : float, default=0.425
        Y-axis position for the center of the `title` text, in relative coordinates. 
        Should be in the range of 0-1. Default is ``0.425`` which, when paired with the 
        default `title_fontsize` and `title` coordinates, will be positioned just under
        the `title`.  

    show_subtitle : bool, default=True
        Whether to show `subtitle`. Only used if `title` is provided, since the default 
        data use to populate `title` and `subtitle` is the same.
             
    width : float, default=0.55  
        Fraction of the donut plot radius that corresponds to the donut 'hole'.  
            
    linewidth : int or float, default=2  
        Width of the lines separating donut segments.  
        
    plot_kwargs : dict, default=None  
        Dictionary containing keyword arguments that will be passed directly to ``ax.pie()``.
            
    text_kwargs : dict, default=None  
        Dictionary containing keyword arguments that will be passed directly to ``ax.text()`` 
        when drawing the title text in the center of the plot.

    subtext_kwargs : dict, default=None  
        Dictionary containing keyword arguments that will be passed directly to ``ax.text()`` 
        when drawing the subtitle text in the center of the plot.

    ax : mpl.axes.Axes, default=None
        Pre-existing axes for the plot. If not provided, a new axes will be created.

    show : bool, default=False  
        If ``True``, plot is shown and the plot ``Axes`` object is not returned. Default
        is ``False``, which does not call ``pyplot.show()`` and returns the ``Axes`` object.

    figsize : iterable object, default=[6, 6]  
        Figure size, in inches. 
            
    figfile : str, optional  
        Path to an output figure file. If not provided, the figure will be shown and not saved to file. 
    

    Returns
    -------
    ax : mpl.axes.Axes
        If `figfile` is ``None`` and `show` is ``False``, the ``ax`` is returned. 
        Otherwise, the return value is ``None``. 
        
    
    .. _Matplotlib color
        https://matplotlib.org/stable/tutorials/colors/colors.html

    """
    # process input data
    d = {}
    # if data is None, then counts/values must be iterables
    if data is None:
        if counts is None:
            counter = Counter(values)
            d["value"] = list(counter.keys())
            d["count"] = list(counter.values())
    # if data isn't None, then we need to determine whether each of
    # counts/values/hue is in data is an iterable, or isn't provided
    else:
        if isinstance(values, str) and values in data.columns:
            values = data[values]
            d["value"] = values
        if counts is None:
            counter = Counter(values)
            d["value"] = list(counter.keys())
            d["count"] = list(counter.values())
        elif isinstance(counts, str) and counts in data.columns:
            d["count"] = data[counts]

    # hue aggregation functions
    most_common_agg_func = lambda x: Counter(x).most_common()[0][0]
    hue_agg_funcs = {
        "max": max,
        "min": min,
        "mean": np.mean,
        "median": np.median,
        "most_common": most_common_agg_func,
    }

    # process hues
    if hue is not None:
        if isinstance(hue, str) and hue in data.columns:
            hue = data[hue]
            hue_vals = {}
            for v, h in zip(values, hue):
                if v not in hue_vals:
                    hue_vals[v] = []
                hue_vals[v].append(h)
            if all([isinstance(h, bool) for h in hue]):
                hue_agg_func = any
            else:
                hue_agg_func = hue_agg_func.get(hue_agg, most_common_agg_func)
            agg_hue_vals = []
            for v in d["value"]:
                h = hue_vals[v]
                if len(h) == 1:
                    agg_hue_vals.append(h[0])
                else:
                    agg_hue_vals.append(hue_agg_func(h))
            d["hue"] = agg_hue_vals
        else:
            d["hue"] = [hue.get(v, None) for v in d["value"]]
    else:
        d["hue"] = d["value"]
    df = pd.DataFrame(d)

    # singletons
    if group_singletons:
        total_count = df.shape[0]
        df = df[df["count"] > 1]
        singleton_count = total_count - df.shape[0]
        singleton_color if singleton_color is not None else alt_color
        singleton_df = pd.DataFrame(
            [
                {
                    "value": "singletons",
                    "count": singleton_count,
                    "order": df.shape[0] + 1,
                    "hue": "singletons",
                    "color": singleton_color,
                },
            ]
        )

    # set the plotting order
    if order is not None:
        order_lookup = {v: i for i, v in enumerate(order)}
        df["order"] = [order_lookup[v] for v in df["value"]]
    else:
        if sort_by in ["hue", "hues"]:
            sort_by = "hue"
        elif sort_by in ["value", "values"]:
            sort_by = "value"
        else:
            sort_by = "count"
        order = sorted(list(set(df[sort_by])), key=sort_key, reverse=sort_descending)
        order_lookup = {v: i for i, v in enumerate(order)}
        df["order"] = [order_lookup[v] for v in df[sort_by]]

    # set hue type
    if force_continuous_hue:
        hue_type = "continuous"
    elif all([isinstance(h, float) for h in df["hue"]]) and not force_categorical_hue:
        hue_type = "continuous"
    elif all([isinstance(h, bool) for h in df["hue"]]):
        hue_type = "boolean"
    else:
        hue_type = "categorical"

    # color
    if hue_type == "continuous":
        if cmap is None:
            if color is not None:
                cmap = get_cmap(color)
            else:
                cmap = get_cmap("flare")
        else:
            cmap = get_cmap(cmap)
        norm_hue = (df["hue"] - df["hue"].min()) / (df["hue"].max() - df["hue"].min())
        df["color"] = [cmap(nh) for nh in norm_hue]
    elif hue_type == "boolean":
        if palette is None:
            pos = color if color is not None else sns.color_palette()[0]
            neg = alt_color
        else:
            pos = palette[True]
            neg = palette[False]
        df["color"] = [pos if h else neg for h in df["hue"]]
    else:
        if palette is None:
            if hue_order is None:
                hue_order = natsorted(df["hue"].dropna().unique())
            if color is not None:
                colors = monochrome_palette(color, n_colors=len(hue_order))
                if shuffle_colors:
                    primary = colors[0]
                    secondary = colors[1:]
                    np.random.seed(random_seed)
                    np.random.shuffle(secondary)
                    colors = [primary] + secondary
            else:
                colors = sns.hls_palette(n_colors=len(hue_order))
                if shuffle_colors:
                    np.random.seed(random_seed)
                    np.random_shuffle(colors)
            palette = {h: c for h, c in zip(hue_order, colors)}
        df["color"] = [palette.get(h, alt_color) for h in df["hue"]]

    # concat the singletons and sort
    if group_singletons:
        df = pd.concat([df, singleton_df], ignore_index=True)
    df = df.sort_values(by="order", ascending=False)

    # make the plot
    if ax is None:
        plt.figure(figsize=figsize if figsize is not None else [6, 6])
        ax = plt.gca()
    ax.axis("equal")
    plot_kws = {"startangle": 90, "radius": 1, "pctdistance": 1 - width / 2}
    if plot_kwargs is not None:
        for k, v in plot_kwargs.items():
            plot_kws[k] = v
    plot_kws["colors"] = df["color"]
    slices, _ = ax.pie(df["count"], **plot_kws)
    plt.setp(slices, width=width, edgecolor=edgecolor)
    for w in slices:
        w.set_linewidth(linewidth)

    # add text to the center of the donut (default is total number of items)
    if title is None:
        show_subtitle = False
        title = str(len(values))
    txt_kws = {
        "size": title_fontsize,
        "color": "k",
        "va": "center",
        "ha": "center",
        "fontweight": "bold",
    }
    if text_kwargs is not None:
        for k, v in text_kwargs.items():
            txt_kws[k] = v
    ax.text(title_x, title_y, title, transform=ax.transAxes, **txt_kws)

    # add the subtitle (default is the total number of items)
    if show_subtitle:
        if subtitle is None:
            subtitle = f"n={len(values)}"
        subtxt_kws = {
            "size": subtitle_fontsize,
            "color": "k",
            "va": "center",
            "ha": "center",
        }
        if subtext_kwargs is not None:
            for k, v in subtext_kwargs.items():
                subtxt_kws[k] = v
        ax.text(subtitle_x, subtitle_y, subtitle, transform=ax.transAxes, **subtxt_kws)

    plt.tight_layout()
    if figfile is not None:
        plt.savefig(figfile)
    elif show:
        plt.show()
    else:
        return ax


# def _get_monochrome_colors(monochrome_color, n_col):
#     cmap = get_cmap(monochrome_color)
#     # this is a bit convoluted, but what's happening is we're getting different colormap
#     # values -- which range from 1 (darkest) to 0 (lightest). Calling cmap(i) returns an
#     # rgba tuple, but we just need the rbg, so we drop the a. To make sure that one of
#     # the colors isn't pure white, we ask np.linspace() for one more value than we need
#     # and drop the lightest value
#     RGB_tuples = [cmap(i)[:-1] for i in np.linspace(1, 0, n_col + 1)][:-1]
#     return RGB_tuples



#!/usr/bin/env python
# filename: mongodb.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# from __future__ import absolute_import, division, print_function, unicode_literals

# import logging
# import platform
# import os
# import subprocess as sp
# import sys

# from pymongo import MongoClient

# from . import log

# if sys.version_info[0] > 2:
#     STR_TYPES = [
#         str,
#     ]
# else:
#     STR_TYPES = [str, unicode]


# def get_connection(ip="localhost", port=27017, user=None, password=None):
#     """
#     Returns a pymongo ``MongoClient`` object.

#     .. note:

#         Both ``user`` and ``password`` are required when connecting to a MongoDB
#         database that has authentication enabled.

#     Arguments:

#         ip (str): IP address of the MongoDB server. Default is ``localhost``.

#         port (int): Port of the MongoDB server. Default is ``27017``.

#         user (str): Username, if authentication is enabled on the MongoDB database.
#             Default is ``None``, which results in requesting the connection
#             without authentication.

#         password (str): Password, if authentication is enabled on the MongoDB database.
#             Default is ``None``, which results in requesting the connection
#             without authentication.
#     """
#     if platform.system().lower() == "darwin":
#         connect = False
#     else:
#         connect = True
#     if user and password:
#         import urllib

#         pwd = urllib.quote_plus(password)
#         uri = "mongodb://{}:{}@{}:{}".format(user, pwd, ip, port)
#         return MongoClient(uri, connect=connect)
#     return MongoClient(ip, port, connect=connect)


# def list_dbs(ip="localhost", port=27017, user=None, connect=True, hint=None):
#     """
#     Returns a list of Databases.

#     .. note:

#         Both ``user`` and ``password`` are required when connecting to a MongoDB
#         database that has authentication enabled.

#     Arguments:

#         db (str): Name of the MongoDB database. Required.

#         ip (str): IP address of the MongoDB server. Default is ``localhost``.

#         port (int): Port of the MongoDB server. Default is ``27017``.

#         user (str): Username, if authentication is enabled on the MongoDB database.
#             Default is ``None``, which results in requesting the connection
#             without authentication.

#         password (str): Password, if authentication is enabled on the MongoDB database.
#             Default is ``None``, which results in requesting the connection
#             without authentication.

#         hint (str): substring found in database name, if used only list of
#             databases that contain the substring will returned
#     """

#     # Darwin says "you shall not connect!"
#     if platform.system().lower() == "darwin":
#         connect = False
#     else:
#         connect = True

#     # check for user and pass, if so: make uri, import urllib, return dbs
#     if user and password:
#         import urllib

#         pwd = urllib.quote_plus(password)
#         uri = "mongodb://{}:{}@{}:{}".format(user, pwd, ip, port)
#         # check to see if hint was passed in and return list of dbs with hint
#         if hint and type(hint) == str:
#             ls = MongoClient(uri, connect=connect).list_database_names()
#             return [d for d in ls if hint.upper() in d.upper()]
#         # otherwise return a list of all dbs
#         else:
#             return MongoClient(uri, connect=connect).list_database_names()

#     # otherwise no user or pass required
#     else:
#         # check to see if hint was passed in and return list of dbs with hint
#         if hint and type(hint) == str:
#             ls = MongoClient(ip, port, connect=connect).list_database_names()
#             return [d for d in ls if hint.upper() in d.upper()]
#         # otherwise return a list of all dbs
#         else:
#             return MongoClient(ip, port, connect=connect).list_database_names()


# def get_db(db, ip="localhost", port=27017, user=None, password=None):
#     """
#     Returns a pymongo ``Database`` object.

#     .. note:

#         Both ``user`` and ``password`` are required when connecting to a MongoDB
#         database that has authentication enabled.

#     Arguments:

#         db (str): Name of the MongoDB database. Required.

#         ip (str): IP address of the MongoDB server. Default is ``localhost``.

#         port (int): Port of the MongoDB server. Default is ``27017``.

#         user (str): Username, if authentication is enabled on the MongoDB database.
#             Default is ``None``, which results in requesting the connection
#             without authentication.

#         password (str): Password, if authentication is enabled on the MongoDB database.
#             Default is ``None``, which results in requesting the connection
#             without authentication.
#     """
#     if platform.system().lower() == "darwin":
#         connect = False
#     else:
#         connect = True
#     if user and password:
#         import urllib

#         pwd = urllib.quote_plus(password)
#         uri = "mongodb://{}:{}@{}:{}".format(user, pwd, ip, port)
#         conn = MongoClient(uri, connect=connect)
#     else:
#         conn = MongoClient(ip, port, connect=connect)
#     return conn[db]


# def get_collections(db, collection=None, prefix=None, suffix=None):
#     """
#     Returns a sorted list of collection names found in ``db``.

#     Arguments:

#         db (Database): A pymongo Database object. Can be obtained
#             with ``get_db``.

#         collection (str): Name of a collection. If the collection is
#             present in the MongoDB database, a single-element list will
#             be returned with the collecion name. If not, an empty list
#             will be returned. This option is primarly included to allow
#             for quick checking to see if a collection name is present.
#             Default is None, which results in this option being ignored.

#         prefix (str): If supplied, only collections that begin with
#             ``prefix`` will be returned.

#         suffix (str): If supplied, only collections that end with
#             ``suffix`` will be returned.

#     Returns:

#         list: A sorted list of collection names.
#     """
#     if collection is not None:
#         return [
#             collection,
#         ]
#     collections = db.collection_names(include_system_collections=False)
#     if prefix is not None:
#         collections = [c for c in collections if c.startswith(prefix)]
#     if suffix is not None:
#         collections = [c for c in collections if c.endswith(suffix)]
#     return sorted(collections)


# def rename_collection(db, collection, new_name):
#     """
#     Renames a MongoDB collection.

#     Arguments:

#         db (Database): A pymongo ``Database`` object. Can be obtained
#             with ``get_db``.

#         collection (str): Name of the collection to be renamed.

#         new_name (str, func): ``new_name`` can be one of two things::

#             1. The new collection name, as a string.
#             2. A function which, when passed the current collection name,
#                 returns the new collection name. If the function
#                 returns an empty string, the collection will not be
#                 renamed.
#     """
#     if hasattr(new_name, "__call__"):
#         _new = new_name(collection)
#         if _new == "":
#             return
#     else:
#         _new = new_name
#     c = db[collection]
#     c.rename(_new)


# def update(field, value, db, collection, match=None):
#     """
#     Updates MongoDB documents.

#     Sets ``field`` equal to ``value`` for all documents that
#     meet ``match`` criteria.

#     Arguments:

#         field (str): Field to update.

#         value (str): Update value.

#         db (Database): A pymongo ``Database`` object.

#         collection (str): Collection name.

#         match (dict): A dictionary containing the match criteria, for example::

#             {'seq_id': {'$in': ['a', 'b', 'c']}, 'cdr3_len': {'$gte': 18}}
#     """
#     c = db[collection]
#     match = match if match is not None else {}
#     # check MongoDB version to use appropriate update command
#     if db.client.server_info()["version"].startswith("2"):
#         c.update(match, {"$set": {field: value}}, multi=True)
#     else:
#         c.update_many(match, {"$set": {field: value}})


# def unset(db, collection, field, match=None):
#     """
#     Removes ``field`` from all records in ``collection`` that meet
#     ``match`` criteria.

#     Arguments:

#         field (str): Field to be removed.

#         db (Database): A pymongo Database object.

#         collection (str): Collection name.

#         match (dict): A dictionary containing the match criteria, for example::

#             {'seq_id': {'$in': ['a', 'b', 'c']}, 'cdr3_len': {'$gte': 18}}
#     """
#     c = db[collection]
#     match = match if match is not None else {}
#     # check MongoDB version to use appropriate update command
#     if db.client.server_info()["version"].startswith("2"):
#         c.update(match, {"$unset": {field: ""}}, multi=True)
#     else:
#         c.update_many(match, {"$unset": {field: ""}})


# def mongoimport(
#     json,
#     database,
#     ip="localhost",
#     port=27017,
#     user=None,
#     password=None,
#     delim="_",
#     delim1=None,
#     delim2=None,
#     delim_occurance=1,
#     delim1_occurance=1,
#     delim2_occurance=1,
# ):
#     """
#     Performs mongoimport on one or more json files.

#     Args:

#         json: Can be one of several things:

#             - path to a single JSON file
#             - an iterable (list or tuple) of one or more JSON file paths
#             - path to a directory containing one or more JSON files

#         database (str): Name of the database into which the JSON files
#             will be imported

#         ip (str): IP address of the MongoDB server. Default is ``localhost``.

#         port (int): Port of the MongoDB database. Default is ``27017``.

#         user (str): Username for the MongoDB database, if authentication is enabled.
#             Default is ``None``, which results in attempting connection without
#             authentication.

#         password (str): Password for the MongoDB database, if authentication is enabled.
#             Default is ``None``, which results in attempting connection without
#             authentication.

#         delim (str): Delimiter, when generating collection names using a single delimiter.
#             Default is ``_``

#         delim_occurance (int): Occurance at which to split filename when using a
#             single delimiter. Default is ``1``

#         delim1 (str): Left delimiter when splitting with two delimiters. Default is None.

#         delim1_occurance (int): Occurance of ``delim1`` at which to split filename.
#             Default is ``1``

#         delim2 (str): Right delimiter when splitting with two delimiters. Default is None.

#         delim2_occurance (int): Occurance of ``delim2`` at which to split filename.
#             Default is ``1``
#     """
#     logger = log.get_logger("mongodb")
#     _print_mongoimport_info(logger)
#     if type(json) in (list, tuple):
#         pass
#     elif os.path.isdir(json):
#         from abtools.utils.pipeline import list_files

#         json = list_files(json)
#     else:
#         json = [
#             json,
#         ]
#     jsons = sorted([os.path.expanduser(j) for j in json if j.endswith(".json")])
#     collections = _get_import_collections(
#         jsons,
#         delim,
#         delim_occurance,
#         delim1,
#         delim1_occurance,
#         delim2,
#         delim2_occurance,
#     )
#     logger.info("Found {} files to import".format(len(jsons)))
#     logger.info("")
#     for i, (json_file, collection) in enumerate(zip(jsons, collections)):
#         logger.info(
#             "[ {} ] {} --> {}".format(i + 1, os.path.basename(json_file), collection)
#         )
#         # logger.info("Performing mongoimport on {}.".format(os.path.basename(json_file)))
#         # logger.info("Importing the file into collection {}.".format(collection))
#         if all([user, password]):
#             host = "--host {} --port {} -username {} -password {}".format(
#                 ip, port, user, password
#             )
#         else:
#             host = "--host {} --port {}".format(ip, port)
#         mongo_cmd = "mongoimport {} --db {} --collection {} --file {}".format(
#             host, database, collection, json_file
#         )
#         mongo = sp.Popen(mongo_cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
#         stdout, stderr = mongo.communicate()


# def index(db, collection, fields, directions=None, desc=False, background=False):
#     """
#     Builds a simple (single field) or complex (multiple fields) index
#     on a single collection in a MongoDB database.

#     Args:

#         db (Database): A pymongo Database object.

#         collection (str): Collection name.

#         fields: Can be one of two things:

#             - the name of a single field, as a string
#             - an iterable (list/tuple) of one or more field names

#         desc (bool): If ``True``, all indexes will be created in descending order.
#             Default is ``False``.

#         directions (list): For complex indexes for which you'd like to have
#             different indexing directions (ascending for some fields, descending
#             for others), you can pass a list of pymongo direction objects (
#             pymongo.ASCENDING and pymongo.DESCENDING), in the same order as the
#             list of fields to be indexed. Must be the same length as the list
#             of index fields. Default is ``None``.

#         background (bool): If ``True``, the indexing operation will be processed
#             in the background. When performing background indexes, the MongoDB
#             database will not be locked.
#     """
#     import pymongo

#     if type(fields) in STR_TYPES:
#         fields = [
#             fields,
#         ]
#     if directions is None:
#         _dir = pymongo.DESCENDING if desc else pymongo.ASCENDING
#         directions = [_dir] * len(fields)
#     field_tuples = list(zip(fields, directions))
#     coll = db[collection]
#     coll.create_index(field_tuples, background=background)


# def remove_padding(db, collection, field="padding"):
#     """
#     Removes a padding field.

#     Args:

#         db (Database): A pymongo Database object.

#         collection (str): Collection name

#         field (str): Name of the padding field. Default is ``padding``
#     """
#     unset(db, collection, field=field)
#     # c = db[collection]
#     # c.update({}, {'$unset': {field: ''}}, multi=True)


# def _get_import_collections(
#     jsons, delim, delim_occurance, delim1, delim1_occurance, delim2, delim2_occurance
# ):
#     jnames = [os.path.basename(j) for j in jsons]
#     if not all([delim1, delim2]):
#         collections = [delim.join(j.split(delim)[:delim_occurance]) for j in jnames]
#     else:
#         pre_colls = [delim1.join(j.split(delim1)[delim1_occurance:]) for j in jnames]
#         collections = [
#             delim2.join(j.split(delim2)[:delim2_occurance]) for j in pre_colls
#         ]
#     return collections


# def _print_mongoimport_info(logger):
#     logger.info("")
#     logger.info("")
#     logger.info("")
#     logger.info("-" * 25)
#     logger.info("MONGOIMPORT")
#     logger.info("-" * 25)
#     logger.info("")


# def _print_remove_padding():
#     logger = log.get_logger("mongodb")
#     logger.info("Removing MongoDB padding...")




import pytest

from ..core.lineage import Lineage, LineageSummary, group_lineages
from ..core.pair import Pair
from ..core.sequence import Sequence


# ==================
#      FIXTURES
# ==================


@pytest.fixture
def mock_heavy_sequences():
    """Create mock heavy chain Sequence objects."""
    return [
        Sequence(
            {
                "sequence_id": f"seq{i}",
                "sequence": "ATCGATCGATCGATCG",
                "sequence_aa": "CASSLDRYFDYW",
                "locus": "IGH",
                "v_call": "IGHV1-2*01",
                "j_call": "IGHJ4*01",
                "c_gene": "IGHG1",
                "cdr3_aa": "CARASGY",
                "cdr3_length": 7,
                "v_identity": 0.95 - (i * 0.01),
                "v_identity_aa": 0.92 - (i * 0.01),
                "lineage": "lineage_001",
            }
        )
        for i in range(5)
    ]


@pytest.fixture
def mock_light_sequences():
    """Create mock light chain Sequence objects."""
    return [
        Sequence(
            {
                "sequence_id": f"seq{i}",
                "sequence": "GCTAGCTAGCTAGCTA",
                "sequence_aa": "CQQSYSTPLT",
                "locus": "IGK",
                "v_call": "IGKV1-5*01",
                "j_call": "IGKJ1*01",
                "cdr3_aa": "CQQSY",
                "cdr3_length": 5,
                "v_identity": 0.97 - (i * 0.01),
                "v_identity_aa": 0.95 - (i * 0.01),
            }
        )
        for i in range(5)
    ]


@pytest.fixture
def mock_pairs(mock_heavy_sequences, mock_light_sequences):
    """Create mock Pair objects for Lineage testing."""
    pairs = []
    for heavy, light in zip(mock_heavy_sequences, mock_light_sequences):
        pairs.append(Pair([heavy, light]))
    return pairs


@pytest.fixture
def mock_pairs_heavy_only(mock_heavy_sequences):
    """Create mock Pair objects with only heavy chains."""
    return [Pair([heavy]) for heavy in mock_heavy_sequences]


@pytest.fixture
def mock_lineage(mock_pairs):
    """Create a mock Lineage object."""
    return Lineage(mock_pairs)


# ================================
#    LINEAGE CLASS BASIC TESTS
# ================================


def test_lineage_init(mock_pairs):
    """Test Lineage initialization."""
    lineage = Lineage(mock_pairs)
    assert lineage is not None
    assert lineage.pairs == mock_pairs


def test_lineage_init_with_name(mock_pairs):
    """Test Lineage initialization with a name."""
    lineage = Lineage(mock_pairs, name="test_lineage")
    assert lineage is not None


def test_lineage_size(mock_lineage):
    """Test Lineage size method."""
    assert mock_lineage.size() == 5


def test_lineage_size_pairs_only(mock_lineage):
    """Test Lineage size method with pairs_only=True."""
    size = mock_lineage.size(pairs_only=True)
    assert size == 5  # All mock pairs have both chains


def test_lineage_size_heavy_only(mock_pairs_heavy_only):
    """Test Lineage size with heavy-only pairs."""
    lineage = Lineage(mock_pairs_heavy_only)
    assert lineage.size() == 5
    assert lineage.size(pairs_only=True) == 0


def test_lineage_contains(mock_lineage, mock_pairs):
    """Test Lineage __contains__ method."""
    # The pair_dict uses pair.name as keys
    pair_name = mock_pairs[0].name
    assert pair_name in mock_lineage


def test_lineage_getitem(mock_lineage, mock_pairs):
    """Test Lineage __getitem__ method."""
    pair_name = mock_pairs[0].name
    retrieved_pair = mock_lineage[pair_name]
    assert retrieved_pair is not None
    assert retrieved_pair.name == pair_name


def test_lineage_iter(mock_lineage, mock_pairs):
    """Test Lineage __iter__ method."""
    pairs_from_iter = list(mock_lineage)
    assert len(pairs_from_iter) == 5
    assert pairs_from_iter == mock_pairs


# ================================
#    LINEAGE PROPERTIES TESTS
# ================================


def test_lineage_heavies_property(mock_lineage):
    """Test Lineage heavies property."""
    heavies = mock_lineage.heavies
    assert len(heavies) == 5
    for pair in heavies:
        assert pair.heavy is not None


def test_lineage_lights_property(mock_lineage):
    """Test Lineage lights property."""
    lights = mock_lineage.lights
    assert len(lights) == 5
    for pair in lights:
        assert pair.light is not None


def test_lineage_just_pairs_property(mock_lineage):
    """Test Lineage just_pairs property."""
    just_pairs = mock_lineage.just_pairs
    assert len(just_pairs) == 5
    for pair in just_pairs:
        assert pair.is_pair


def test_lineage_name_property(mock_lineage):
    """Test Lineage name property returns string or None."""
    name = mock_lineage.name
    # Name can be None if lineage info isn't found in expected format
    assert name is None or isinstance(name, str)


def test_lineage_pair_dict_property(mock_lineage, mock_pairs):
    """Test Lineage pair_dict property."""
    pair_dict = mock_lineage.pair_dict
    assert isinstance(pair_dict, dict)
    assert len(pair_dict) == 5
    for pair in mock_pairs:
        assert pair.name in pair_dict


def test_lineage_has_insertion(mock_lineage):
    """Test Lineage has_insertion property."""
    # Our mock sequences don't have insertions
    assert mock_lineage.has_insertion is False


def test_lineage_has_deletion(mock_lineage):
    """Test Lineage has_deletion property."""
    # Our mock sequences don't have deletions
    assert mock_lineage.has_deletion is False


def test_lineage_has_indel(mock_lineage):
    """Test Lineage has_indel property."""
    # Our mock sequences don't have indels
    assert mock_lineage.has_indel is False


# ================================
#    LINEAGE SUMMARY TESTS
# ================================


def test_lineage_summary_init_with_pairs(mock_pairs):
    """Test LineageSummary initialization with Pair objects."""
    summary = LineageSummary(mock_pairs, name="test_summary")
    assert summary is not None
    assert summary.name == "test_summary"
    assert len(summary.pairs) == 5


def test_lineage_summary_heavies_lights(mock_pairs):
    """Test LineageSummary heavies and lights attributes."""
    summary = LineageSummary(mock_pairs, name="test_summary")
    assert len(summary.heavies) == 5
    assert len(summary.lights) == 5


def test_lineage_summary_df(mock_pairs):
    """Test LineageSummary has a df attribute."""
    summary = LineageSummary(mock_pairs, name="test_summary")
    assert summary.df is not None
    assert len(summary.df) == 5


# ================================
#    GROUP LINEAGES TESTS
# ================================


def test_group_lineages_basic():
    """Test group_lineages function."""
    # Create pairs with clonify IDs
    pairs = []
    for i in range(5):
        heavy = Sequence(
            {
                "sequence_id": f"seq{i}",
                "sequence": "ATCG",
                "locus": "IGH",
                "clonify": {"id": "lineage_001" if i < 3 else "lineage_002"},
            }
        )
        light = Sequence(
            {
                "sequence_id": f"seq{i}",
                "sequence": "GCTA",
                "locus": "IGK",
            }
        )
        pairs.append(Pair([heavy, light]))

    lineages = group_lineages(pairs)
    assert len(lineages) == 2

    # Check lineage sizes
    sizes = sorted([l.size() for l in lineages], reverse=True)
    assert sizes[0] == 3
    assert sizes[1] == 2


def test_group_lineages_just_pairs():
    """Test group_lineages with just_pairs=True."""
    # Create pairs with clonify IDs, some without light chains
    pairs = []
    for i in range(5):
        heavy = Sequence(
            {
                "sequence_id": f"seq{i}",
                "sequence": "ATCG",
                "locus": "IGH",
                "clonify": {"id": "lineage_001"},
            }
        )
        if i < 3:  # Only first 3 have light chains
            light = Sequence(
                {
                    "sequence_id": f"seq{i}",
                    "sequence": "GCTA",
                    "locus": "IGK",
                }
            )
            pairs.append(Pair([heavy, light]))
        else:
            pairs.append(Pair([heavy]))

    # Without just_pairs
    lineages_all = group_lineages(pairs, just_pairs=False)
    assert len(lineages_all) == 1
    assert lineages_all[0].size() == 5

    # With just_pairs
    lineages_pairs = group_lineages(pairs, just_pairs=True)
    assert len(lineages_pairs) == 1
    assert lineages_pairs[0].size() == 3


def test_group_lineages_empty():
    """Test group_lineages with no clonify annotations."""
    pairs = []
    for i in range(3):
        heavy = Sequence(
            {
                "sequence_id": f"seq{i}",
                "sequence": "ATCG",
                "locus": "IGH",
                # No clonify annotation
            }
        )
        pairs.append(Pair([heavy]))

    lineages = group_lineages(pairs)
    assert len(lineages) == 0



# from .core import *
# from .utils import *

import os

from . import bin, cl, io, log, pl, tl
from . import cl as color
from .core import lineage, pair, sequence
from .core.lineage import Lineage
from .core.pair import Pair
from .core.sequence import Sequence

# from .tools import log
# from .tools.phylo import Phylogeny
# from .utils import (
#     alignment,
#     cluster,
#     # color,
#     decorators,
#     jobs,
#     path,
#     # log,
#     phylogeny,
#     progbar,
#     s3,
#     seqio,
#     utilities,
# )
# from .utils import alignment as aln
# from .utils import path as path
# # from .utils.alignment import SSWAlignment, NWAlignment
# from .tools.alignment import GlobalAlignment, LocalAlignment, SemiGlobalAlignment
# from .utils.jobs import *
# from .utils.progbar import progress_bar
from .version import __version__

BINARY_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "binaries"))



#!/usr/bin/env python
# filename: ridge.py


#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import warnings

warnings.filterwarnings("ignore")

import itertools
from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt

from natsort import natsorted


def ridge(
    categories: Union[str, Iterable, None] = None,
    values: Union[str, Iterable, None] = None,
    data: Optional[pd.DataFrame] = None,
    order: Optional[Iterable] = None,
    palette: Union[dict, Iterable, None] = None,
    alt_color: Union[Iterable, str] = "lightgrey",
    alpha: float = 1.0,
    linewidth: float = 0.0,
    outlinewidth: float = 1.5,
    xaxis_linewidth: float = 1.0,
    xaxis_linecolor: Union[Iterable, str, None] = None,
    xlabel: str = "UMI count ($\mathregular{log_2}$)",
    xlabel_fontsize: int = 12,
    ylabel_fontsize: int = 11,
    bw: Union[str, float] = "scott",
    adjust_hspace: float = 0.1,
    category_label_fontweight: str = "bold",
    category_label_xoffset: float = 0.25,
    xmin: Union[int, float, None] = None,
    xmax: Union[int, float, None] = None,
    aspect: float = 15.0,
    height: float = 0.5,
    figfile: Optional[str] = None,
    show: bool = True,
):
    """
    Draws a ridge plot.

    Parameters
    ----------
    categories : Union[str, Iterable, None], optional
        Category information. Can be any of the following:
          - if `data` is provided and `values` is not provided, `categories`
            must be a list containing one or more column names in `data`
          - if both `data` and `values` are also provided, `categories` should be a
            column name in `data`
          - if `data` is not provided, `categories` must be a list of category names

    values : Union[str, Iterable, None], optional
        Value information. Can be any of the following:
          - if `data` and `categories` are provided, `values` should be a column name in `data`
          - if `data` is not provided, `values` must be a list of values
        Note that `values` cannot be provided without `categories`.

    data : Optional[pd.DataFrame], optional
        Input ``DataFrame``. If not provided, ``categories`` and ``values`` must be provided.

    order : Optional[Iterable], optional
        Order of categories in the plot. If not provided, categories will be sorted using ``natsort``.

    palette : Union[dict, Iterable, None], optional
        Color palette. Can be any of the following:
            - a dictionary mapping categories to colors
            - a list of colors
        If a ``dict`` is provided, any categories not included in the dictionary will be assigned
        `alt_color`. If a ``list`` is provided, the colors will be assigned to the categories in order,
        with colors reused if there are more categories than colors. If not provided, a default palette
        (created using ``sns.hls_palette()``) will be used.

    alt_color : Union[Iterable, str], optional
        Color to use for categories not included in the palette. If a ``list`` is provided, the color

    alpha : float, optional
        Alpha value for the density curves. Default is 1.0.

    linewidth : float, optional
        Line width for the data line on the density curves. Default is 0.0.

    outlinewidth : float, optional
        Line width for the white outline of the density curves. The purpose is to provide visual
        separation if/when the density curves of adjacent ridge plots overlap. Default is 1.5.

    xaxis_linewidth : float, optional
        Line width for the x-axis line. Default is 1.0.

    xaxis_linecolor : Union[Iterable, str, None], optional
        Color for the x-axis line. If not provided, the color will be the same as the color
        used to plot the density curve.

    xlabel : str, optional
        Label for the x-axis. Default is "UMI count ($\mathregular{log_2}$)".

    xlabel_fontsize : int, optional
        Font size for the x-axis label. Default is 12.

    ylabel_fontsize : int, optional
        Font size for the y-axis labels. Default is 11.

    bw : Union[str, float], optional
        Bandwidth for the density curves. Can be any of the following:
            - a float
            - the name of a reference rule (see ``scipy.stats.gaussian_kde``)
        Default is "scott".

    adjust_hspace : float, optional
        Adjust the vertical space between subplots. Default is 0.1.

    category_label_fontweight : str, optional
        Font weight for the category labels. Default is "bold".

    category_label_xoffset : float, optional
        Horizontal offset for the category labels, as a fraction of the total plot width.
        Default is 0.25.

    xmin : Union[int, float, None], optional
        Minimum value for the x-axis. If not provided, the minimum value will be the data minimum.

    xmax : Union[int, float, None], optional
        Maximum value for the x-axis. If not provided, the maximum value will be the data maximum.

    aspect : float, optional
        Aspect ratio for the plot. Default is 15.0.

    height : float, optional
        Height of each individual ridge plot, in inches. Default is 0.5.

    figfile : Optional[str], optional
        If provided, the figure will be saved to this file. Default is None.

    show : bool, optional
        If True, the figure will be displayed. Default is True.

    Returns
    -------
    sns.FacetGrid
        If `show` is ``False`` and `figfile` is not provided, the ``FacetGrid`` object
        will be returned.

    """
    # process input data
    if data is None:
        # if just `categories` and `values` are provided, we create a dataframe
        d = {}
        d["category"] = categories
        d["values"] = values
        df = pd.DataFrame(d)
        categories = "category"
        values = "values"
    elif all([data is not None, categories is not None, values is None]):
        # if `data` is provided without `values`, `categories` must be a list of column names
        # we melt the data to get the dataframe into the correct format
        df = data.copy()
        if not all([c in df.columns for c in categories]):
            raise ValueError(
                f"Invalid column names for categories: {categories}. "
                f"If `data` is provided without `values`, `categories` must be a list of column names. "
                f"Available columns: {df.columns}"
            )
        df = df[categories]
        df = df.melt(value_vars=categories, var_name="category", value_name="values")
        categories = "category"
        values = "values"
    else:
        # if `data` is provided with `values` and `categories`, we check to see if they are column names
        # if not (and if the length matches the number of rows in `data`), we create new columns
        df = data.copy()
        if not isinstance(categories, str) and len(categories) == df.shape[0]:
            df["category"] = categories
            categories = "category"
        if not isinstance(values, str) and len(values) == df.shape[0]:
            df["values"] = values
            values = "values"

    # categories and order
    if order is None:
        order = natsorted(df[categories].unique())
    n_categories = len(order)
    df = df.query(f"{categories} in @order")

    # colors
    if palette is None:
        palette = {
            cat: color
            for cat, color in zip(order, sns.hls_palette(n_categories, l=0.5, s=0.8))
        }
    elif palette is not None and isinstance(palette, (list, tuple)):
        iter_colors = itertools.cycle(palette)
        palette = {cat: color for cat, color in zip(order, iter_colors)}
    elif palette is not None and isinstance(palette, dict):
        palette = {cat: palette.get(cat, alt_color) for cat in order}
    else:
        raise ValueError(
            f"Invalid palette: {palette}. Must be an iterable of colors or dict of category: color pairs."
        )
    df["color"] = df[categories].map(palette)

    # make the plot
    g = sns.FacetGrid(
        df,
        row=categories,
        hue=categories,
        aspect=aspect,
        height=height,
        palette=palette,
        row_order=order,
        hue_order=order,
    )
    g.map(
        sns.kdeplot, values, clip_on=False, shade=True, alpha=alpha, lw=linewidth, bw=bw
    )
    g.map(sns.kdeplot, values, clip_on=False, color="w", lw=outlinewidth, bw=bw)
    g.map(plt.axhline, y=0, lw=xaxis_linewidth, color=xaxis_linecolor, clip_on=False)

    # add labels to the individual ridge plots
    def label(x, color, label):
        ax = plt.gca()
        ax.text(
            0,
            0.3,
            label,
            fontweight=category_label_fontweight,
            color=color,
            fontsize=ylabel_fontsize,
            ha="left",
            va="center",
            transform=ax.transAxes,
        )

    g.map(label, categories)

    # style the plot
    if xmin is None:
        xmin = df[values].min()
    if xmax is None:
        xmax = df[values].max()
    category_label_xoffset = (xmax - xmin) * category_label_xoffset
    g.fig.subplots_adjust(hspace=adjust_hspace)
    g.set_titles("")
    g.set(xticks=range(int(np.floor(xmin,)), int(np.ceil(xmax)) + 1, 2,))
    g.set(xlim=(xmin - category_label_xoffset, xmax + 0.25))
    g.set(yticks=[])
    g.despine(bottom=True, left=True)

    # set and locate the X axis label
    g.set(ylabel="")
    g.set(xlabel=xlabel)
    xlabel_position = (((xmax - xmin) / 2) + category_label_xoffset) / (
        (xmax - xmin) + category_label_xoffset
    )
    for ax in g.axes.flat:
        ax.set_xlabel(ax.get_xlabel(), x=xlabel_position, fontsize=xlabel_fontsize)

    # save or show the plot
    if figfile:
        g.savefig(figfile)
    elif show:
        plt.show()
    return g



from collections import Counter

import pytest

from ..core.sequence import Sequence
from ..tools.clonify import clonify, pairwise_distance


@pytest.fixture
def clonify_seqs():
    return [
        Sequence(
            dict(
                sequence_id="seq1",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq2",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq3",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq4",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq5",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq6",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq7",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq8",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-2",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq9",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-69",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
        Sequence(
            dict(
                sequence_id="seq10",
                sequence="ATCGATCGATCG",
                v_gene="IGHV1-69",
                j_gene="IGHJ4",
                cdr3="CASSLDRYFDYW",
                v_mutations="",
                locus="IGH",
            )
        ),
    ]


@pytest.fixture
def pairwise_distance_seqs():
    seq1 = Sequence(
        dict(
            sequence_id="seq1",
            sequence="ATCGATCGATCG",
            v_gene="IGHV1-69",
            j_gene="IGHJ4",
            cdr3="CASSLDRYFDYW",
            mutations=["V1A", "V2G", "V3C"],
        )
    )
    seq2 = Sequence(
        dict(
            sequence_id="seq2",
            sequence="ATCGATCGATCG",
            v_gene="IGHV1-69",
            j_gene="IGHJ4",
            cdr3="CASSLDRYFDYW",
            mutations=["V1A", "V2G", "V3C"],
        )
    )
    seq3 = Sequence(
        dict(
            sequence_id="seq3",
            sequence="ATCGATCGATCG",
            v_gene="IGHV1-69",
            j_gene="IGHJ4",
            cdr3="CASSLDYFDYW",
            mutations=["V1A", "V2G", "V3T"],
        )
    )
    return seq1, seq2, seq3


def test_clonify(clonify_seqs):
    results = clonify(clonify_seqs, return_assignment_dict=True)
    # sanity checks on the results
    assert isinstance(results, dict)
    for seq in clonify_seqs:
        assert seq.id in results
    # check for correct lineage sizes
    lineage_sizes = Counter(results.values())
    lineage_sizes = dict(
        sorted(lineage_sizes.items(), key=lambda x: x[1], reverse=True)
    )
    assert len(lineage_sizes) == 2
    lineage_names = list(lineage_sizes.keys())
    assert lineage_sizes[lineage_names[0]] == 8
    assert lineage_sizes[lineage_names[1]] == 2


def test_pairwise_distance(pairwise_distance_seqs):
    seq1, seq2, seq3 = pairwise_distance_seqs
    dist1 = pairwise_distance(seq1, seq2)
    dist2 = pairwise_distance(seq1, seq3)
    assert isinstance(dist1, float)
    assert isinstance(dist2, float)
    assert dist1 <= 0.01
    assert dist2 > 0.0



#!/usr/bin/python
# filename:  codons.py


#
# Copyright (c) 2015 Bryan Briney
# License:  The MIT license (http: //opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


codon_lookup = {
    "TTT": "F",
    "TTC": "F",
    "TTA": "L",
    "TTG": "L",
    "TCT": "S",
    "TCC": "S",
    "TCA": "S",
    "TCG": "S",
    "TAT": "Y",
    "TAC": "Y",
    "TAA": "*",
    "TAG": "*",
    "TGT": "C",
    "TGC": "C",
    "TGA": "*",
    "TGG": "W",
    "CTT": "L",
    "CTC": "L",
    "CTA": "L",
    "CTG": "L",
    "CCT": "P",
    "CCC": "P",
    "CCA": "P",
    "CCG": "P",
    "CAT": "H",
    "CAC": "H",
    "CAA": "Q",
    "CAG": "Q",
    "CGT": "R",
    "CGC": "R",
    "CGA": "R",
    "CGG": "R",
    "ATT": "I",
    "ATC": "I",
    "ATA": "I",
    "ATG": "M",
    "ACT": "T",
    "ACC": "T",
    "ACA": "T",
    "ACG": "T",
    "AAT": "N",
    "AAC": "N",
    "AAA": "K",
    "AAG": "K",
    "AGT": "S",
    "AGC": "S",
    "AGA": "R",
    "AGG": "R",
    "GTT": "V",
    "GTC": "V",
    "GTA": "V",
    "GTG": "V",
    "GCT": "A",
    "GCC": "A",
    "GCA": "A",
    "GCG": "A",
    "GAT": "D",
    "GAC": "D",
    "GAA": "E",
    "GAG": "E",
    "GGT": "G",
    "GGC": "G",
    "GGA": "G",
    "GGG": "G",
}
